/* Copyright (C) 1979-1999 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.
   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* Calculate a checksum for a row */

#include "myisamdef.h"

ha_checksum mi_checksum(MI_INFO *info, const byte *buf)
{
  uint i;
  ha_checksum crc=0;
  MI_COLUMNDEF *rec=info->s->rec;

  for (i=info->s->base.fields ; i-- ; buf+=(rec++)->length)
  {
    const byte *pos;
    const byte *end;
    switch (rec->type) {
    case FIELD_BLOB:
    {
      ulong length=_mi_calc_blob_length(rec->length-
					mi_portable_sizeof_char_ptr,
					buf);
      memcpy((char*) &pos, buf+rec->length- mi_portable_sizeof_char_ptr,
	     sizeof(char*));
      end=pos+length;
      break;
    }
    case FIELD_VARCHAR:
    {
      uint length;
      length=uint2korr(buf);
      pos=buf+2; end=pos+length;
      break;
    }
    default:
      pos=buf; end=buf+rec->length;
      break;
    }
    for ( ; pos != end ; pos++)
      crc=((crc << 8) + *((uchar*) pos)) + (crc >> (8*sizeof(ha_checksum)-8));
  }
  return crc;
}


ha_checksum mi_static_checksum(MI_INFO *info, const byte *pos)
{
  ha_checksum crc;
  const byte *end=pos+info->s->base.reclength;
  for (crc=0; pos != end; pos++)
      crc=((crc << 8) + *((uchar*) pos)) + (crc >> (8*sizeof(ha_checksum)-8));
  return crc;
}
