/* Copyright (C) 1999 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.
   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* Read and write key blocks */

#include "myisamdef.h"
#ifdef	__WIN32__
#include <errno.h>
#endif

	/* Fetch a key-page in memory */

uchar *_mi_fetch_keypage(register MI_INFO *info, MI_KEYDEF *keyinfo,
			 my_off_t page, uchar *buff, int return_buffer)
{
  uchar *tmp;
  tmp=(uchar*) key_cache_read(info->s->kfile,page,(byte*) buff,
			     (uint) keyinfo->block_length,
			     (uint) keyinfo->block_length,
			     return_buffer);
  if (tmp == info->buff)
  {
    info->update|=HA_STATE_BUFF_SAVED;
    info->int_pos=page;
  }
  else
  {
    info->update&= ~HA_STATE_BUFF_SAVED;
    if (tmp)
      info->int_pos=page;
    else
    {
      info->int_pos=HA_OFFSET_ERROR;
      DBUG_PRINT("error",("Got errno: %d from key_cache_read",my_errno));
      my_errno=HA_ERR_CRASHED;
    }
  }
  return tmp;
} /* _mi_fetch_keypage */


	/* Write a key-page on disk */

int _mi_write_keypage(register MI_INFO *info, register MI_KEYDEF *keyinfo,
		      my_off_t page, uchar *buff)
{
  reg3 uint length;
#ifndef FAST					/* Safety check */
  if (page < info->s->base.keystart ||
      page+keyinfo->block_length > info->s->state.key_file_length ||
      page & (myisam_block_size-1))
  {
    DBUG_PRINT("error",("Trying to write outside key region: %lu",
			(long) page));
    my_errno=EINVAL;
    return(-1);
  }
  DBUG_PRINT("page",("write page at: %lu",(long) page,buff));
  DBUG_DUMP("buff",(byte*) buff,mi_getint(buff));
#endif

  if ((length=keyinfo->block_length) > IO_SIZE*2 &&
       info->s->state.key_file_length != page+length)
    length= ((mi_getint(buff)+IO_SIZE-1) & (uint) ~(IO_SIZE-1));
#ifdef HAVE_purify
  {
    length=mi_getint(buff);
    bzero((byte*) buff+length,keyinfo->block_length-length);
    length=keyinfo->block_length;
  }
#endif
  return (key_cache_write(info->s->kfile,page,(byte*) buff,length,
			 (uint) keyinfo->block_length,
			 (int) (info->lock_type != F_UNLCK)));
} /* mi_write_keypage */


	/* Remove page from disk */

int _mi_dispose(register MI_INFO *info, MI_KEYDEF *keyinfo, my_off_t pos)
{
  my_off_t old_link;
  char buff[8];
  DBUG_ENTER("_mi_dispose");

  old_link=info->s->state.key_del[keyinfo->block_size];
  info->s->state.key_del[keyinfo->block_size]=pos;
  mi_sizestore(buff,old_link);
  DBUG_RETURN(key_cache_write(info->s->kfile,pos,buff,
			      sizeof(buff),
			      (uint) keyinfo->block_length,
			      (int) (info->lock_type != F_UNLCK)));
} /* _mi_dispose */


	/* Make new page on disk */

my_off_t _mi_new(register MI_INFO *info, MI_KEYDEF *keyinfo)
{
  my_off_t pos;
  char buff[8];
  DBUG_ENTER("_mi_new");

  if ((pos=info->s->state.key_del[keyinfo->block_size]) == HA_OFFSET_ERROR)
  {
    if (info->s->state.key_file_length >=
	info->s->base.max_key_file_length - keyinfo->block_length)
    {
      my_errno=HA_ERR_INDEX_FILE_FULL;
      DBUG_RETURN(HA_OFFSET_ERROR);
    }
    pos=info->s->state.key_file_length;
    info->s->state.key_file_length+= keyinfo->block_length;
  }
  else
  {
    if (!key_cache_read(info->s->kfile,pos,
			buff,
			(uint) sizeof(buff),
			(uint) keyinfo->block_length,0))
      pos= HA_OFFSET_ERROR;
    else
      info->s->state.key_del[keyinfo->block_size]=mi_sizekorr(buff);
  }
  DBUG_PRINT("exit",("Pos: %d",pos));
  DBUG_RETURN(pos);
} /* _mi_new */
