/* Copyright (C) 1979-1999 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.
   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* Descript, check and repair of ISAM tables */

#include "myisamdef.h"

#include <m_ctype.h>
#include <stdarg.h>
#include <getopt.h>
#ifdef HAVE_SYS_VADVICE_H
#include <sys/vadvise.h>
#endif
#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif
SET_STACK_SIZE(9000)			/* Minimum stack size for program */

#define T_VERBOSE	1
#define T_SILENT	2
#define T_DESCRIPT	4
#define T_EXTEND	8
#define T_INFO		16
#define T_REP		32
#define T_OPT		64		/* Not currently used */
#define T_FORCE_CREATE	128
#define T_WRITE_LOOP	256
#define T_UNPACK	512
#define T_STATISTICS	1024
#define T_VERY_SILENT	2048
#define T_SORT_RECORDS	4096
#define T_SORT_INDEX	8192
#define T_WAIT_FOREVER	16384
#define T_REP_BY_SORT	32768L
#define T_FAST		65536L
#define T_READONLY	131072L
#define T_MEDIUM	T_READONLY*2
#define T_AUTO_INC	T_MEDIUM*2
#define T_CHECK		T_AUTO_INC*2

#define O_NEW_INDEX	1		/* Bits set in out_flag */
#define O_NEW_DATA	2

#if defined(_MSC_VER) && !defined(__WIN32__)
#define USE_BUFFER_INIT		250L*1024L
#define READ_BUFFER_INIT	((uint) 32768-MALLOC_OVERHEAD)
#define SORT_BUFFER_INIT	(uint) (65536L-MALLOC_OVERHEAD)
#define MIN_SORT_BUFFER		(1024*16-MALLOC_OVERHEAD)
#else
#define USE_BUFFER_INIT		(((1024L*512L-MALLOC_OVERHEAD)/IO_SIZE)*IO_SIZE)
#define READ_BUFFER_INIT	(1024L*256L-MALLOC_OVERHEAD)
#define SORT_BUFFER_INIT	(2048L*1024L-MALLOC_OVERHEAD)
#define MIN_SORT_BUFFER		(4096-MALLOC_OVERHEAD)
#endif

#define NEAD_MEM	((uint) 10*4*(IO_SIZE+32)+32) /* Nead for recursion */
#define MAXERR			20
#define BUFFERS_WHEN_SORTING	16		/* Alloc for sort-key-tree */
#define WRITE_COUNT		MY_HOW_OFTEN_TO_WRITE
#define INDEX_TMP_EXT		".TMM"
#define DATA_TMP_EXT		".TMD"
#define MYF_RW	MYF(MY_NABP | MY_WME | MY_WAIT_IF_FULL)

#define UPDATE_TIME		1
#define UPDATE_STAT		2
#define UPDATE_SORT		4
#define UPDATE_AUTO_INC		8

typedef struct st_sort_key_blocks {		/* Used when sorting */
  uchar *buff,*end_pos;
  uchar lastkey[MI_MAX_POSSIBLE_KEY_BUFF];
  uint last_length;
  int inited;
} SORT_KEY_BLOCKS;

typedef struct st_sort_info {
  MI_INFO *info;
  enum data_file_type new_data_file_type;
  SORT_KEY_BLOCKS *key_block,*key_block_end;
  uint key,find_length;
  my_off_t pos,max_pos,filepos,start_recpos,filelength,dupp,max_records,
    buff_length;
  ulonglong unique[MI_MAX_KEY_SEG+1];
  my_bool fix_datafile;
  char *record,*buff;
  MI_KEYDEF *keyinfo;
  MI_KEYSEG *keyseg;
} SORT_INFO;


static ulong	use_buffers=0,read_buffer_length=0,write_buffer_length=0,
		sort_buffer_length=0,sort_key_blocks=0,crc=0;
static uint out_flag=0,warning_printed=0,error_printed=0,
	    opt_rep_quick=0,verbose=0,opt_follow_links=1;
static uint opt_sort_key=0,total_files=0,max_level=0;
static ulong testflag=0,decode_bits;
static ulonglong keys_in_use= ~(ulonglong) 0;
static ha_rows total_records=0,total_deleted=0;
static my_off_t new_file_pos=0,key_file_blocks=0;
static my_off_t search_after_block=HA_OFFSET_ERROR;
static my_off_t keydata=0,totaldata=0,key_blocks=0;
static ha_checksum record_checksum=0;
static ulonglong auto_increment_value= (ulonglong) ~0;
static ulonglong max_data_file_length=0;
static my_string type_names[]={ "?","char","binary", "short", "long", "float",
				"double","number","unsigned short",
				"unsigned long","longlong","ulonglong","int24",
				"uint24","int8","varchar", "varbin","?",
				"?"},
	      prefix_packed_txt="packed ",
	      bin_packed_txt="prefix ",
	      diff_txt="stripped ",
	      null_txt="NULL",
	      blob_txt="BLOB ",
	      field_pack[]={"","no endspace", "no prespace",
			    "no zeros", "blob", "constant", "table-lockup",
			    "always zero","varchar","unique-hash","?","?"};

static char temp_filename[FN_REFLEN],*isam_file_name,*tmpdir;
static IO_CACHE read_cache;
static SORT_INFO sort_info;
static int tmpfile_createflag=O_RDWR | O_TRUNC | O_EXCL;
static ulonglong unique_count[MI_MAX_KEY_SEG+1];
static ha_checksum key_crc[MI_MAX_POSSIBLE_KEY];
static char **default_argv;

static my_string load_default_groups[]= { "myisamchk", "isamchk",0 };

	/* Functions defined in this file */

extern int main(int argc,char * *argv);
extern void print_error _VARARGS((char *fmt,...));
static void print_warning _VARARGS((char *fmt,...));
static void print_info _VARARGS((char *fmt,...));
static int myisamchk(char *filename);
static void get_options(int *argc,char * * *argv);
static int chk_del(MI_INFO *info,uint testflag);
static int check_k_link(MI_INFO *info,uint nr);
static int chk_size(MI_INFO *info);
static int chk_key(MI_INFO *info);
static int chk_index(MI_INFO *info,MI_KEYDEF *keyinfo,my_off_t page,
		     uchar *buff, ha_rows *keys, ha_checksum *key_checksum,
		     uint level);
static uint isam_key_length(MI_INFO *info,MI_KEYDEF *keyinfo);
static ha_checksum calc_checksum(ha_rows count);
static int chk_data_link(MI_INFO *info,int extend);
static int rep(MI_INFO *info,char *name, int rep_quick);
static int writekeys(MI_INFO *info,byte *buff,my_off_t filepos);
static void descript(MI_INFO *info,char *name);
static int movepoint(MI_INFO *info,byte *record,my_off_t oldpos,my_off_t newpos,
		     uint prot_key);
static void lock_memory(void);
static int flush_blocks(File file);
static int sort_records(MI_INFO *,my_string,uint,int);
static int sort_index(MI_INFO *info,my_string name);
static int sort_record_index(MI_INFO *info,MI_KEYDEF *keyinfo,my_off_t page,
			     uchar *buff,uint sortkey,File new_file);
static int sort_one_index(MI_INFO *info,MI_KEYDEF *keyinfo,uchar *buff,
			  File new_file);
static int change_to_newfile(my_string filename,my_string old_ext,
			     my_string new_ext);
static int lock_file(File file,my_off_t start,int lock_type,my_string filetype,
		     my_string filename);
static int filecopy(File to,File from,my_off_t start,my_off_t length,
		    my_string type);

static void print_version(void);
static int rep_by_sort(MI_INFO *info,my_string name, int rep_quick);
static int sort_key_read(void *key);
static int sort_get_next_record(void);
static int sort_write_record(void);
static int sort_key_cmp(const void *a,const void *b);
static int sort_key_write(const void *a);
static my_off_t get_record_for_key(MI_INFO *info,MI_KEYDEF *keyinfo,
				uchar *key);
static int sort_insert_key(reg1 SORT_KEY_BLOCKS *key_block,uchar *key,
			   my_off_t prev_block);
static int sort_delete_record(void);
static void usage(void);
static int flush_pending_blocks(void);
static SORT_KEY_BLOCKS	*alloc_key_blocks(uint blocks,uint buffer_length);
static int test_if_almost_full(MI_INFO *info);
static int recreate_database(MI_INFO **info,char *filename);
static int write_data_suffix(MI_INFO *info);
static int update_state_info(MI_INFO *info,uint update);
char *llstr(longlong value,char *buff);
static void update_auto_increment_key(MI_INFO *info);
static void update_key_parts(MI_KEYDEF *keyinfo,
			     ulong *rec_per_key_part,
			     ulonglong *unique,
			     ulonglong records);
static ha_checksum mi_byte_checksum(const byte *buf, uint length);

#ifdef __WIN32__
static double ulonglong2double(ulonglong value)
{
  longlong nr=(longlong) value;
  if (nr >= 0)
    return (double) nr;
  return (18446744073709551616.0 + (double) nr);
}
#endif

	/* Main program */

int main(int argc, char **argv)
{
  int error;
  MY_INIT(argv[0]);

  get_options(&argc,(char***) &argv);
  myisam_quick_table_bits=(uint) decode_bits;
  error=0;
  while (--argc >= 0)
  {
    error|= myisamchk(*(argv++));
    VOID(fflush(stdout));
    VOID(fflush(stderr));
    if ((error_printed | warning_printed) && (testflag & T_FORCE_CREATE) &&
	(!(testflag & (T_REP | T_REP_BY_SORT | T_SORT_RECORDS |
		       T_SORT_INDEX))))
    {
      testflag|=T_REP;
      error|=myisamchk(argv[-1]);
      testflag&= ~T_REP;
      VOID(fflush(stdout));
      VOID(fflush(stderr));
    }
    if (argc && (!(testflag & T_SILENT) || testflag & T_INFO))
    {
      puts("\n---------\n");
      VOID(fflush(stdout));
    }
  }
  if (total_files > 1)
  {					/* Only if descript */
    char buff[22],buff2[22];
    if (!(testflag & T_SILENT) || testflag & T_INFO)
      puts("\n---------\n");
    printf("\nTotal of all %d ISAM-files:\nData records: %9s   Deleted blocks: %9s\n",total_files,llstr(total_records,buff),llstr(total_deleted,buff2));
  }
  free_defaults(default_argv);
  my_end(testflag & T_INFO ? MY_CHECK_ERROR | MY_GIVE_INFO : MY_CHECK_ERROR);
  exit(error);
#ifndef _lint
  return 0;				/* No compiler warning */
#endif
} /* main */


static CHANGEABLE_VAR changeable_vars[] = {
  { "key_buffer_size",(long*) &use_buffers,(long) USE_BUFFER_INIT,
    (long) MALLOC_OVERHEAD, (long) ~0L,(long) MALLOC_OVERHEAD,(long) IO_SIZE },
  { "read_buffer_size", (long*) &read_buffer_length,(long) READ_BUFFER_INIT,
      (long) MALLOC_OVERHEAD,(long) ~0L,(long) MALLOC_OVERHEAD,(long) 1L },
  { "write_buffer_size", (long*) &write_buffer_length,(long) READ_BUFFER_INIT,
      (long) MALLOC_OVERHEAD,(long) ~0L,(long) MALLOC_OVERHEAD,(long) 1L },
  { "sort_buffer_size",(long*) &sort_buffer_length,(long) SORT_BUFFER_INIT,
      (long) (MIN_SORT_BUFFER+MALLOC_OVERHEAD),(long) ~0L,
      (long) MALLOC_OVERHEAD,(long) 1L },
  { "sort_key_blocks",(long*) &sort_key_blocks,BUFFERS_WHEN_SORTING,4L,100L,0L,
    1L },
  { "decode_bits",(long*) &decode_bits,9L,4L,17L,0L,1L },
  { NullS,(long*) 0,0L,0L,0L,0L,0L,} };


static struct option long_options[] =
{
  {"analyze",	       no_argument,	  0, 'a'},
  {"block-search",     required_argument, 0, 'b'},
  {"check",	       no_argument,	  0, 'c'},
#ifndef DBUG_OFF
  {"debug",	       required_argument, 0, '#'},
#endif
  {"description",      no_argument,	  0, 'd'},
  {"data-file-length", required_argument, 0, 'D'},
  {"extend-check",     no_argument,	  0, 'e'},
  {"fast",	       no_argument,	  0, 'F'},
  {"force",	       no_argument,	  0, 'f'},
  {"help",	       no_argument,	  0, '?'},
  {"information",      no_argument,	  0, 'i'},
  {"keys-used",        required_argument, 0, 'k'},
  {"medium-check",     no_argument,	  0, 'm'},
  {"no-symlinks",      no_argument,	  0, 'l'},
  {"quick",	       no_argument,	  0, 'q'},
  {"read-only",        no_argument,	  0, 'T'},
  {"recover",	       no_argument,	  0, 'r'},
  {"safe-recover",     no_argument,	  0, 'o'},
  {"set-auto-increment",optional_argument, 0, 'A'},
  {"set-variable",     required_argument, 0, 'O'},
  {"silent",	       no_argument,	  0, 's'},
  {"sort-index",       no_argument,	  0, 'S'},
  {"sort-records",     required_argument, 0, 'R'},
  {"tmpdir",	       required_argument, 0, 't'},
  {"unpack",	       no_argument,	  0, 'u'},
  {"verbose",	       no_argument,	  0, 'v'},
  {"version",	       no_argument,	  0, 'V'},
  {"wait",	       no_argument,	  0, 'w'},
  {0, 0, 0, 0}
};

static void print_version(void)
{
  printf("%s  Ver 1.0 for %s at %s\n",my_progname,SYSTEM_TYPE,
	 MACHINE_TYPE);
}

static void usage(void)
{
  uint i;
  print_version();
  puts("TCX Datakonsult AB, by Monty, for your professional use");
  puts("This software comes with NO WARRANTY: see the PUBLIC for details.\n");
  puts("Description, check and repair of ISAM tables.");
  puts("Used without options all tables on the command will be checked for errors");
  printf("Usage: %s [OPTIONS] tables[.MYI]\n", my_progname);
  puts("\nGlobal options:\n\
  -#, --debug=...     Output debug log. Often this is 'd:t:o,filename`\n\
  -?, --help          Display this help and exit.\n\
  -O, --set-variable var=option\n\
		      Change the value of a variable.\n\
  -s, --silent	      Only print errors.  One can use two -s to make\n\
		      myisamchk very silent\n\
  -v, --verbose       Print more information. This can be used with\n\
                      --describe and --check. Use many -v for more verbosity!\n\
  -V, --version       Print version and exit.\n\
  -w, --wait          Wait if table is locked.\n");

  puts("Check options (check is the default action for myisamchk):\n\
  -c, --check	      Check table for errors\n\
  -e, --extend-check  Check the table VERY throughly.  One need use this\n\
                      only in extreme cases as myisamchk should normally find\n\
                      all errors even without this switch\n\
  -F, --fast	      Don't check tables that has not changed since last check\n\
  -f, --force         Restart with -r if there are any errors in the table\n\
  -i, --information   Print statistics information about table that is checked\n\
  -m, --medium-check  Faster than extended-check, but only finds 99.99% of\n\
		      all errors.  Should however be good enough for most cases\n\
  -T, --read-only     Don't mark table as checked\n");

  puts("Repair options (When using -r or -o) \n\
  -D, --data-file-length=#  Max length of data file (when recreating data\n\
                      file when it's full)\n\
  -f, --force         Overwrite old temporary files.\n\
  -k, --keys-used=#   Tell MY_ISAM to update only the first\n\
		      # keys.  This can be used to get faster inserts!\n\
  -l, --no-symlinks   Do not follow symbolic links. Normally\n\
		      myisamchk repairs the table a symlink points at.\n\
  -r, --recover       Can fix almost anything except unique keys that aren't\n\
                      unique.\n\
  -o, --safe-recover  Uses old recovery method; Slower than '-r' but can\n\
		      handle a couple of cases where '-r' reports that it can't\n\
		      fix the data file.\n\
  -t, --tmpdir=path   Path for temporary files\n\
  -q, --quick         Faster repair by not modifying the data file.\n\
                      One can give a second '-q' to force myisamchk to\n\
		      modify the original datafile in case of duplicate keys\n\
  -u, --unpack        Unpack file packed with pack_isam.\n\
");

  puts("Other actions:\n\
  -a, --analyze	      Analyze distribution of keys. Will make some joins in\n\
		      MySQL faster.  You can check the calculated distribution\n\
		      by using '--describe --verbose table_name'.\n\
  -d, --description   Prints some information about table.\n\
  -A, --set-auto-increment[=value]\n\
		      Force auto_increment to start at this or higher value\n\
		      If no value is given, then sets the next auto increment\n\
		      value to the higest used value for the auto key + 1.\n\
  -S, --sort-index    Sort index blocks.  This speeds up 'read-next' in\n\
		      applications\n\
  -R, --sort-records=#\n\
		      Sort records according to an index.  This makes your\n\
		      data much more localized and may speed up things\n\
		      (It may be VERY slow to do a sort the first time!)\n");

  printf("Possible variables for option --set-variable (-O) are:\n");
  for (i=0; changeable_vars[i].name ; i++)
    printf("%-20s  current value: %u\n",
	   changeable_vars[i].name,
	   *changeable_vars[i].varptr);
}

	/* Check table */

static int myisamchk(my_string filename)
{
  int error,lock_type,recreate;
  MI_INFO *info;
  File datafile;
  char fixed_name[FN_REFLEN];
  char llbuff[22],llbuff2[22];
  MYISAM_SHARE *share;
  int rep_quick=opt_rep_quick;
  DBUG_ENTER("myisamchk");

  out_flag=error=warning_printed=error_printed=recreate=0;
  datafile=0;
  isam_file_name=filename;			/* For error messages */
  if (!(info=mi_open(filename,
		     (testflag & (T_DESCRIPT | T_READONLY)) ? O_RDONLY : O_RDWR,
		     (testflag & T_WAIT_FOREVER) ? HA_OPEN_WAIT_IF_LOCKED :
		     (testflag & T_DESCRIPT) ? HA_OPEN_IGNORE_IF_LOCKED :
		     HA_OPEN_ABORT_IF_LOCKED)))
  {
    /* Avoid twice printing of isam file name */
    error_printed=1;
    switch (my_errno) {
    case HA_ERR_CRASHED:
      print_error("'%s' is not a MYISAM-table",filename);
      break;
    case HA_ERR_OLD_FILE:
      print_error("'%s' is a old type of MYISAM-table", filename);
      break;
    case HA_ERR_END_OF_FILE:
      print_error("Couldn't read complete header from '%s'", filename);
      break;
    case EAGAIN:
      print_error("'%s' is locked. Use -w to wait until unlocked",filename);
      break;
    case ENOENT:
      print_error("File '%s' doesn't exist",filename);
      break;
    case EACCES:
      print_error("You don't have permission to use '%s'",filename);
      break;
    default:
      print_error("%d when opening MYISAM-table '%s'",
		  my_errno,filename);
      break;
    }
    DBUG_RETURN(1);
  }
  share=info->s;
  share->options&= ~HA_OPTION_READ_ONLY_DATA; /* We are modifing it */

  /* If run with --fast, don't do anything if file hasn't changed */
  if (testflag & T_FAST && !share->state.changed)
  {
    if (!(testflag & T_SILENT) || testflag & T_INFO)
      printf("MYISAM file: %s is already checked\n",filename);
    if (mi_close(info))
    {
      print_error("%d when closing MYISAM-table '%s'",my_errno,filename);
      DBUG_RETURN(1);
    }
    DBUG_RETURN(0);
  }

  if ((testflag & (T_REP_BY_SORT | T_REP | T_STATISTICS |
		   T_SORT_RECORDS | T_SORT_INDEX)) &&
      ((testflag & T_UNPACK && share->data_file_type == COMPRESSED_RECORD) ||
       mi_uint2korr(share->state.header.state_info_length) !=
       MI_STATE_INFO_SIZE ||
       mi_uint2korr(share->state.header.base_info_length) !=
       MI_BASE_INFO_SIZE ||
       ((keys_in_use & ~ share->state.key_map) &
	((ulonglong) 1L << share->base.keys)-1) ||
       test_if_almost_full(info) ||
       info->s->state.header.file_version[3] != myisam_file_magic[3]))
  {
    if (recreate_database(&info,filename))
    {
      VOID(fprintf(stderr,
		   "MYISAM-table '%s' is not fixed because of errors\n",
	      filename));
      return(-1);
    }
    recreate=1;
    if (!(testflag & (T_REP | T_REP_BY_SORT)))
    {
      testflag|=T_REP_BY_SORT;			/* if only STATISTICS */
      if (!(testflag & T_SILENT))
	printf("- '%s' has old table-format. Recreating index\n",filename);
      if (!rep_quick)
	rep_quick=1;
    }
    share=info->s;
  }

  if (testflag & T_DESCRIPT)
  {
    total_files++;
    total_records+=share->state.records; total_deleted+=share->state.del;
    descript(info,filename);
  }
  else
  {
    if (!(testflag & T_READONLY))
      lock_type = F_WRLCK;			/* table is changed */
    else
      lock_type= F_RDLCK;
    if (info->lock_type == F_RDLCK)
      info->lock_type=F_UNLCK;			/* Read only table */
    if (_mi_readinfo(info,lock_type,0))
    {
      print_error("Can't lock indexfile of '%s', error: %d",
		  filename,my_errno);
      error_printed=0;
      goto end;
    }
    if (lock_file(info->dfile,0L,lock_type,"datafile of",filename))
      goto end;
    info->lock_type= F_EXTRA_LCK;		/* Simulate as locked */
    info->tmp_lock_type=lock_type;
    datafile=info->dfile;

    if (testflag & (T_REP+T_REP_BY_SORT+T_SORT_RECORDS+T_SORT_INDEX))
    {
      if (testflag & (T_REP+T_REP_BY_SORT))
	share->state.key_map= (((ulonglong) 1 << share->base.keys)-1)
	  & keys_in_use;
      VOID(fn_format(fixed_name,filename,"",MI_NAME_IEXT,
		     4+ (opt_follow_links ? 16 : 0)));

      if (rep_quick && chk_del(info,testflag & ~T_VERBOSE))
      {
	if (testflag & T_FORCE_CREATE)
	{
	  rep_quick=0;
	  print_info("Creating new data file\n");
	}
	else
	{
	  error=1;
	  print_error("Quick-recover aborted; Run recovery without switch 'q'");
	}
      }
      if (!error)
      {
	if (testflag & T_REP_BY_SORT &&
	    (share->state.key_map || (rep_quick && !keys_in_use && !recreate)))
	  error=rep_by_sort(info,fixed_name,rep_quick);
	else if (testflag & (T_REP | T_REP_BY_SORT))
	  error=rep(info,fixed_name,rep_quick);
      }
      if (!error && testflag & T_SORT_RECORDS)
      {
	if (out_flag & O_NEW_DATA)
	{			/* Change temp file to org file */
	  VOID(lock_file(datafile,0L,F_UNLCK,"datafile of",filename));
	  VOID(my_close(datafile,MYF(MY_WME)));    /* Close old file */
	  VOID(my_close(info->dfile,MYF(MY_WME))); /* Close new file */
	  error|=change_to_newfile(fixed_name,MI_NAME_DEXT,DATA_TMP_EXT);
	  if ((info->dfile=my_open(fn_format(temp_filename,fixed_name,"",
					     MI_NAME_DEXT,2+4),
				   O_RDWR | O_SHARE,
				   MYF(MY_WME))) <= 0 ||
	      lock_file(info->dfile,0L,F_WRLCK,"datafile",temp_filename))
	    error=1;
	  out_flag&= ~O_NEW_DATA; /* We are using new datafile */
	  read_cache.file=info->dfile;
	}
	if (! error)
	  error=sort_records(info,fixed_name,opt_sort_key,
			     test(!(testflag & T_REP)));
	datafile=info->dfile;	/* This is now locked */
      }
      if (!error && testflag & T_SORT_INDEX)
	error=sort_index(info,fixed_name);
      if (!error)
	share->state.changed=0;
      else
	share->state.changed|=2;
    }
    else if ((testflag & T_CHECK) || !(testflag & T_AUTO_INC))
    {
      if (!(testflag & T_SILENT) || testflag & T_INFO)
	printf("Checking MYISAM file: %s\n",filename);
      if (!(testflag & T_SILENT))
	printf("Data records: %7s   Deleted blocks: %7s\n",
	       llstr(share->state.records,llbuff),
	       llstr(share->state.del,llbuff2));
      if (share->state.open_count)
      {
	print_warning("%d clients is using or hasn't closed the table properly",
		      share->state.open_count);
      }
      share->state.key_map &=keys_in_use;
      error =chk_size(info);
      if (!error || !(testflag & (T_FAST | T_FORCE_CREATE)))
	error|=chk_del(info,testflag);
      if (!error || !(testflag & (T_FAST | T_FORCE_CREATE)))
	error|=chk_key(info);
      if (!rep_quick && !error || !(testflag & (T_FAST | T_FORCE_CREATE)))
      {
	if (testflag & (T_EXTEND | T_MEDIUM))
	  VOID(init_key_cache(use_buffers,(uint) NEAD_MEM));
	VOID(init_io_cache(&read_cache,datafile,(uint) read_buffer_length,
			  READ_CACHE,share->pack.header_length,1,
			  MYF(MY_WME)));
	lock_memory();
	if ((info->s->options & (HA_OPTION_PACK_RECORD |
				 HA_OPTION_COMPRESS_RECORD)) ||
	    (testflag & (T_EXTEND | T_MEDIUM)))
	  error|=chk_data_link(info,testflag & T_EXTEND);
	error|=flush_blocks(share->kfile);
	VOID(end_io_cache(&read_cache));
      }
      if (!error)
      {
	if (share->state.changed)
	{
	  share->state.changed=0;
	  info->update|=HA_STATE_CHANGED;
	}
      }
      else if (!(share->state.changed & 2))
      {						/* Mark crashed */
	share->state.changed|=2;
	info->update|=HA_STATE_CHANGED;
      }
    }
  }
  if (testflag & T_AUTO_INC)
    update_auto_increment_key(info);

end:
  if (!(testflag & T_DESCRIPT))
  {
    if (info->update & HA_STATE_CHANGED && ! (testflag & T_READONLY))
      error|=update_state_info(info,
			       ((testflag & T_REP | T_REP_BY_SORT) ?
			       UPDATE_TIME | UPDATE_STAT : 0) |
			       ((testflag & T_SORT_RECORDS) ?
				UPDATE_SORT : 0));
    VOID(lock_file(share->kfile,0L,F_UNLCK,"indexfile",filename));
    if (datafile > 0)
      VOID(lock_file(datafile,0L,F_UNLCK,"datafile of",filename));
    info->update&= ~HA_STATE_CHANGED;
  }

  if (datafile && datafile != info->dfile)
    VOID(my_close(datafile,MYF(MY_WME)));
  if (mi_close(info))
  {
    print_error("%d when closing MYISAM-table '%s'",my_errno,filename);
    DBUG_RETURN(1);
  }
  if (error == 0)
  {
    if (out_flag & O_NEW_DATA)
      error|=change_to_newfile(fixed_name,MI_NAME_DEXT,DATA_TMP_EXT);
    if (out_flag & O_NEW_INDEX)
      error|=change_to_newfile(fixed_name,MI_NAME_IEXT,INDEX_TMP_EXT);
  }
  VOID(fflush(stdout)); VOID(fflush(stderr));
  if (error_printed)
  {
    if (testflag & (T_REP+T_REP_BY_SORT+T_SORT_RECORDS+T_SORT_INDEX))
      VOID(fprintf(stderr,
		   "MYISAM-table '%s' is not fixed because of errors\n",
		   filename));
    else if (! (error_printed & 2) && !(testflag & T_FORCE_CREATE))
      VOID(fprintf(stderr,
      "MYISAM-table '%s' is corrupted\nFix it using switch \"-r\" or \"-o\"\n",
	      filename));
  }
  else if (warning_printed &&
	   ! (testflag & (T_REP+T_REP_BY_SORT+T_SORT_RECORDS+T_SORT_INDEX+
			  T_FORCE_CREATE)))
    VOID(fprintf(stderr, "MYISAM-table '%s' is usable but should be fixed\n",
	    filename));
  VOID(fflush(stderr));
  DBUG_RETURN(error);
} /* myisamchk */


	 /* Read options */

static void get_options(register int *argc,register char ***argv)
{
  int c,version,option_index=0;
  uint old_testflag;
  char *pos;

  load_defaults("my",load_default_groups,argc,argv);
  default_argv= *argv;
  set_all_changeable_vars(changeable_vars);
  if (isatty(fileno(stdout)))
    testflag|=T_WRITE_LOOP;
  while ((c=getopt_long(*argc,*argv,"acdeif?lqrmosSTuvVw#:b:D:k:O:R:A::t:",
			long_options, &option_index)) != EOF)
  {
    version=0;
    switch(c) {
    case 'a':
      testflag|= T_STATISTICS;
      break;
    case 'A':
      if (optarg)
	auto_increment_value=strtoull(optarg,NULL,0);
      else
	auto_increment_value=0;		/* Set to min used value */
      testflag|= T_AUTO_INC;
      break;
    case 'b':
      search_after_block=strtoul(optarg,NULL,10);
      break;
    case 'c':
      testflag|= T_CHECK;
      break;
    case 'D':
      max_data_file_length=strtoll(optarg,NULL,10);
      break;
    case 's':				/* silent */
      if (testflag & T_SILENT)
	testflag|=T_VERY_SILENT;
      testflag|= T_SILENT;
      testflag&= ~T_WRITE_LOOP;
      break;
    case 'w':
      testflag|= T_WAIT_FOREVER;
      break;
    case 'd':				/* description if isam-file */
      testflag|= T_DESCRIPT;
      break;
    case 'e':				/* extend check */
      testflag|= T_EXTEND;
      break;
    case 'i':
      testflag|= T_INFO;
      break;
    case 'f':
      tmpfile_createflag= O_RDWR | O_TRUNC;
      testflag|=T_FORCE_CREATE;
      break;
    case 'F':
      testflag|=T_FAST;
      break;
    case 'k':
      keys_in_use= (ulonglong) strtoll(optarg,NULL,10);
      break;
    case 'l':
      opt_follow_links=0;
      break;
    case 'm':
      testflag|= T_MEDIUM;		/* Medium check */
      break;
    case 'r':				/* Repair table */
      testflag= (testflag & ~T_REP) | T_REP_BY_SORT;
      break;
    case 'o':
      testflag= (testflag & ~T_REP_BY_SORT) | T_REP;
      my_disable_async_io=1;		/* More safety */
      break;
    case 'q':
      opt_rep_quick++;
      break;
    case 'u':
      testflag|= T_UNPACK | T_REP_BY_SORT;
      break;
    case 'v':				/* Verbose */
      testflag|= T_VERBOSE;
      verbose++;
      break;
    case 'O':
      if (set_changeable_var(optarg, changeable_vars))
      {
	usage();
	exit(1);
      }
      break;
    case 'R':				/* Sort records */
      old_testflag=testflag;
      testflag|= T_SORT_RECORDS;
      opt_sort_key=(uint) atoi(optarg)-1;
      if (opt_sort_key >= MI_MAX_KEY)
      {
	fprintf(stderr,
		"The value of the sort key is bigger than max key: %d.\n",
		MI_MAX_KEY);
	exit(1);
      }
      break;
    case 'S':			      /* Sort index */
      old_testflag=testflag;
      testflag|= T_SORT_INDEX;
      break;
    case 't':
      tmpdir=optarg;
      break;
    case 'T':
      testflag|= T_READONLY;
    case '#':
      DBUG_PUSH(optarg ? optarg : "d:t:o,/tmp/isamchk");
      pos=" ";				/* Skip rest of arg */
      break;
    case 'V':
      version=1;
      print_version();
      exit(0);
    case '?':
      usage();
      exit(0);
    }
  }
  (*argc)-=optind;
  (*argv)+=optind;
  if (*argc == 0)
  {
    usage();
    exit(-1);
  }
  if ((testflag & T_UNPACK) && (opt_rep_quick || (testflag & T_SORT_RECORDS)))
  {
    VOID(fprintf(stderr,
		 "%s: --unpack can't be used with --quick or --sort-records\n",
		 my_progname));
    exit(1);
  }
  if ((testflag & T_READONLY) &&
      (testflag & (T_REP_BY_SORT | T_REP | T_STATISTICS | T_AUTO_INC |
		   T_SORT_RECORDS | T_SORT_INDEX)))
  {
    VOID(fprintf(stderr,
		 "%s: Can't use --readonly when repairing or sorting\n",
		 my_progname));
    exit(1);
  }
  return;
} /* get options */


	/* Check delete links */

static int chk_del(register MI_INFO *info, uint test_flag)
{
  reg2 ha_rows i;
  uint j,delete_link_length;
  my_off_t empty,next_link,old_link;
  char buff[22],buff2[22];
  DBUG_ENTER("chk_del");

  if (!(test_flag & T_SILENT))
    puts("- check key delete-chain");

  LINT_INIT(old_link);
  record_checksum=0;
  key_file_blocks=info->s->base.keystart;
  for (j=0 ; j < info->s->state.header.max_block_size ; j++)
    if (check_k_link(info,j))
      goto wrong;
  delete_link_length=((info->s->options & HA_OPTION_PACK_RECORD) ? 20 :
		      info->s->rec_reflength+1);

  if (!(test_flag & T_SILENT))
    puts("- check record delete-chain");

  next_link=info->s->state.dellink;
  if (info->s->state.del == 0)
  {
    if (test_flag & T_VERBOSE)
    {
      puts("No recordlinks");
    }
  }
  else
  {
    if (test_flag & T_VERBOSE)
      printf("Recordlinks:    ");
    empty=0;
    for (i= info->s->state.del ; i > 0L && next_link != HA_OFFSET_ERROR ; i--)
    {
      if (test_flag & T_VERBOSE)
	printf(" %9s",llstr(next_link,buff));
      if (next_link >= info->s->state.data_file_length)
	goto wrong;
      if (my_pread(info->dfile,(char*) buff,delete_link_length,
		   next_link,MYF(MY_NABP)))
      {
	if (test_flag & T_VERBOSE) puts("");
	print_error("Can't read delete-link at filepos: %s",
		    llstr(next_link,buff));
	DBUG_RETURN(1);
      }
      if (*buff != '\0')
      {
	if (test_flag & T_VERBOSE) puts("");
	print_error("Record at pos: %s is not remove-marked",
		    llstr(next_link,buff));
	goto wrong;
      }
      if (info->s->options & HA_OPTION_PACK_RECORD)
      {
	my_off_t prev_link=mi_sizekorr(buff+12);
	if (empty && prev_link != old_link)
	{
	  if (test_flag & T_VERBOSE) puts("");
	  print_error("Deleted block at %s doesn't point back at previous delete link",llstr(next_link,buff2));
	  goto wrong;
	}
	old_link=next_link;
	next_link=mi_sizekorr(buff+4);
	empty+=mi_uint3korr(buff+1);
      }
      else
      {
	record_checksum+=next_link;
	next_link=_mi_rec_pos(info->s,buff+1);
	empty+=info->s->base.pack_reclength;
      }
    }
    if (empty != info->s->state.empty)
    {
      if (test_flag & T_VERBOSE) puts("");
      print_warning("Not used space is supposed to be: %s but is: %s",
		    llstr(info->s->state.empty,buff),
		    llstr(empty,buff2));
      info->s->state.empty=empty;
    }
    if (i != 0 || next_link != HA_OFFSET_ERROR)
      goto wrong;

    if (test_flag & T_VERBOSE) puts("\n");
  }
  DBUG_RETURN(0);
wrong:
  if (test_flag & T_VERBOSE) puts("");
  print_error("record delete-link-chain corrupted");
  DBUG_RETURN(1);
} /* chk_del */


	/* Check delete links in index file */

static int check_k_link(register MI_INFO *info, uint nr)
{
  my_off_t next_link,block_size=(nr+1)*MI_KEY_BLOCK_LENGTH;
  ha_rows records;
  char llbuff[21];
  DBUG_ENTER("check_k_link");

  if (testflag & T_VERBOSE)
    printf("block_size %4d:",block_size);

  next_link=info->s->state.key_del[nr];
  records= (info->s->state.key_file_length / block_size);
  while (next_link != HA_OFFSET_ERROR && records > 0)
  {
    if (testflag & T_VERBOSE) printf("%16s",llstr(next_link,llbuff));
    if (next_link > info->s->state.key_file_length ||
	next_link & (info->s->blocksize-1))
      DBUG_RETURN(1);
    if (my_pread(info->s->kfile,llbuff,8,next_link, MYF(MY_NABP)))
      DBUG_RETURN(1);
    next_link=mi_sizekorr(llbuff);
    records--;
    key_file_blocks+=block_size;
  }
  if (testflag & T_VERBOSE)
    if (next_link != HA_OFFSET_ERROR)
      printf("%16s\n",llstr(next_link,llbuff));
    else
      puts("");
  DBUG_RETURN (next_link != HA_OFFSET_ERROR);
} /* check_k_link */


	/* Kontrollerar storleken p} filerna */

static int chk_size(register MI_INFO *info)
{
  int error=0;
  register my_off_t skr,size;
  char buff[22],buff2[22];
  DBUG_ENTER("chk_size");

  if (!(testflag & T_SILENT)) puts("- check file-size");

  size=my_seek(info->s->kfile,0L,MY_SEEK_END,MYF(0));
  if ((skr=(my_off_t) info->s->state.key_file_length) != size)
  {
    if (skr > size)
    {
      error=1;
      print_error("Size of indexfile is: %-8s        Should be: %s",
		  llstr(size,buff), llstr(skr,buff2));
    }
    else
      print_warning("Size of indexfile is: %-8s      Should be: %s",
		    llstr(size,buff), llstr(skr,buff2));
  }
  if (!(testflag & T_VERY_SILENT) &&
      ! (info->s->options & HA_OPTION_COMPRESS_RECORD) &&
      info->s->state.key_file_length >
      (ulong) (ulong_to_double(info->s->base.max_key_file_length)*0.9))
    print_warning("Keyfile is almost full, %10s of %10s used",
		  llstr(info->s->state.key_file_length,buff),
		  llstr(info->s->base.max_key_file_length-1,buff));

  size=my_seek(info->dfile,0L,MY_SEEK_END,MYF(0));
  skr=(my_off_t) info->s->state.data_file_length;
  if (info->s->options & HA_OPTION_COMPRESS_RECORD)
    skr+= MEMMAP_EXTRA_MARGIN;
#ifdef USE_RELOC
  if (info->data_file_type == STATIC_RECORD &&
      skr < (my_off_t) info->s->base.reloc*info->s->base.min_pack_length)
    skr=(my_off_t) info->s->base.reloc*info->s->base.min_pack_length;
#endif
  if (skr != size)
  {
    info->s->state.data_file_length=size;	/* Skipp other errors */
    if (skr > size && skr != size + MEMMAP_EXTRA_MARGIN)
    {
      error=1;
      print_error("Size of datafile is: %-8s         Should be: %s",
		    llstr(size,buff), llstr(skr,buff2));
    }
    else
    {
      print_warning("Size of datafile is: %-8s       Should be: %s",
		    llstr(size,buff), llstr(skr,buff2));

    }
  }
  if (!(testflag & T_VERY_SILENT) &&
      !(info->s->options & HA_OPTION_COMPRESS_RECORD) &&
      info->s->state.data_file_length >
      (ulong) (ulong_to_double(info->s->base.max_data_file_length)*0.9))
    print_warning("Datafile is almost full, %10s of %10s used",
		  llstr(info->s->state.data_file_length,buff),
		  llstr(info->s->base.max_data_file_length-1,buff2));
  DBUG_RETURN(error);
} /* chk_size */


	/* Check keys */

static int chk_key(register MI_INFO *info)
{
  uint key,found_keys=0;
  ha_rows keys;
  ha_checksum old_record_checksum,init_checksum;
  my_off_t all_keydata,all_totaldata,key_totlength,length;
  ulong   *rec_per_key_part;
  MYISAM_SHARE *share=info->s;
  MI_KEYDEF *keyinfo;
  char buff[22],buff2[22];
  DBUG_ENTER("chk_key");

  if (!(testflag & T_SILENT)) puts("- check index reference");

  all_keydata=all_totaldata=key_totlength=0;
  old_record_checksum=0;
  init_checksum=record_checksum;
  if (!(share->options &
	(HA_OPTION_PACK_RECORD | HA_OPTION_COMPRESS_RECORD)))
    old_record_checksum=calc_checksum(share->state.records+share->state.del-1)*
      share->base.pack_reclength;
  rec_per_key_part= info->s->state.rec_per_key_part;
  for (key= 0,keyinfo= &share->keyinfo[0]; key < share->base.keys ;
       rec_per_key_part+=keyinfo->keysegs, key++, keyinfo++)
  {
    key_crc[key]=0;
    if (!(((ulonglong) 1 << key) & share->state.key_map))
      continue;
    found_keys++;

    record_checksum=init_checksum;
    bzero((char*) &unique_count,sizeof(unique_count));
    if ((!(testflag & T_SILENT)))
      printf ("- check data record references index: %d\n",key+1);
    if (share->state.key_root[key] == HA_OFFSET_ERROR &&
	share->state.records == 0)
      continue;
    if (!_mi_fetch_keypage(info,keyinfo,share->state.key_root[key],info->buff,
			   0))
    {
      print_error("Can't read indexpage from filepos: %s",
		  llstr(share->state.key_root[key],buff));
      DBUG_RETURN(-1);
    }
    key_file_blocks+=keyinfo->block_length;
    keys=keydata=totaldata=key_blocks=0; max_level=0;
    if (chk_index(info,keyinfo,share->state.key_root[key],info->buff,&keys,
		  key_crc+key,1))
      DBUG_RETURN(-1);
    if (keys != share->state.records)
    {
      print_error("Found %s keys of %s",llstr(keys,buff),
		  llstr(share->state.records,buff2));
      DBUG_RETURN(-1);
    }
    if (!key && (share->options &
		 (HA_OPTION_PACK_RECORD | HA_OPTION_COMPRESS_RECORD)))
      old_record_checksum=record_checksum;
    else if (old_record_checksum != record_checksum)
    {
      if (key)
	print_error("Key %u doesn't point at same records that key 1",
		    key+1);
      else
	print_error("Key 1 doesn't point at all records");
      DBUG_RETURN(-1);
    }
    /* Check that auto_increment key is bigger than max key value */
    if ((uint) share->base.auto_key -1 == key)
    {
      ulonglong save_auto_value=info->s->state.auto_increment;
      info->s->state.auto_increment=0;
      _mi_read_key_record(info, 0L, info->rec_buff);
      update_auto_increment(info, info->rec_buff);
      if (info->s->state.auto_increment > save_auto_value)
      {
	print_warning("Auto-increment value: %s is smaller than max used value: %s",
		      llstr(save_auto_value,buff2),
		      llstr(info->s->state.auto_increment, buff));
      }
      if (testflag & T_AUTO_INC)
      {
	set_if_bigger(info->s->state.auto_increment,auto_increment_value);
      }
      else
	info->s->state.auto_increment=save_auto_value;
    }

    length=(my_off_t) isam_key_length(info,keyinfo)*keys + key_blocks*2;
    if (testflag & T_INFO && totaldata != 0L && keys != 0L)
      printf("Key: %2d:  Keyblocks used: %3d%%  Packed: %4d%%  Max levels: %2d\n",
	     key+1,
	     (int) (keydata*100.0/totaldata),
	     (int) (((double) length - (double) keydata)*100.0/
		    (double) length),
	     max_level);
    all_keydata+=keydata; all_totaldata+=totaldata; key_totlength+=length;

    update_key_parts(keyinfo, rec_per_key_part, unique_count,
		     (ulonglong) share->state.records);
  }
  if (testflag & T_INFO)
  {
    if (all_totaldata != 0L && found_keys > 0)
      printf("Total:    Keyblocks used: %3d%%  Packed: %4d%%\n\n",
	     (int) (all_keydata*100.0/all_totaldata),
	     (int) (((double) key_totlength- (double) all_keydata)*100.0/
		    (double) key_totlength));
    else if (all_totaldata != 0L && share->state.key_map)
      puts("");
  }
  if (key_file_blocks != share->state.key_file_length)
    print_warning("Some data are unreferenced in keyfile");
  record_checksum-=init_checksum;		/* Remove delete links */
  if (testflag & (T_STATISTICS | T_AUTO_INC))
    DBUG_RETURN(update_state_info(info,UPDATE_STAT));
  DBUG_RETURN(0);
} /* chk_key */


	/* Check if index is ok */

static int chk_index(MI_INFO *info, MI_KEYDEF *keyinfo, my_off_t page,
		     uchar *buff, ha_rows *keys, ha_checksum *key_checksum,
		     uint level)
{
  int flag;
  uint used_length,comp_flag,nod_flag,key_length,not_used;
  uchar key[MI_MAX_POSSIBLE_KEY_BUFF],*temp_buff,*keypos,*endpos;
  my_off_t next_page,record;
  char llbuff[22],llbuff2[22];
  DBUG_ENTER("chk_index");
  DBUG_DUMP("buff",(byte*) buff,mi_getint(buff));

  if (!(temp_buff=(uchar*) my_alloca((uint) keyinfo->block_length)))
  {
    print_error("Not Enough memory");
    DBUG_RETURN(-1);
  }

  if (keyinfo->flag & HA_NOSAME)
    comp_flag=SEARCH_FIND | SEARCH_UPDATE;	/* Not real dupplicates */
  else
    comp_flag=SEARCH_SAME;			/* Keys in positionorder */
  nod_flag=mi_test_if_nod(buff);
  used_length=mi_getint(buff);
  keypos=buff+2+nod_flag;
  endpos=buff+used_length;

  keydata+=used_length; totaldata+=keyinfo->block_length;	/* INFO */
  key_blocks++;
  if (level > max_level)
    max_level=level;

  if (used_length > keyinfo->block_length)
  {
    print_error("Wrong pageinfo at page: %s", llstr(page,llbuff));
    goto err;
  }
  for ( ;; )
  {
    if (nod_flag)
    {
      next_page=_mi_kpos(nod_flag,keypos);
      if (next_page > info->s->state.key_file_length ||
	  (nod_flag && (next_page & (info->s->blocksize -1))))
      {
	my_off_t max_length=my_seek(info->s->kfile,0L,MY_SEEK_END,MYF(0));
	print_error("Wrong pagepointer: %s at page: %s",
		    llstr(next_page,llbuff),llstr(page,llbuff2));

	if (next_page+info->s->blocksize > max_length)
	  goto err;
	info->s->state.key_file_length=(max_length &
					~ (my_off_t) (info->s->blocksize-1));
      }
      if (!_mi_fetch_keypage(info,keyinfo,next_page,temp_buff,0))
      {
	print_error("Can't read key from filepos: %s",llstr(next_page,llbuff));
	goto err;
      }
      key_file_blocks+=keyinfo->block_length;
      if (chk_index(info,keyinfo,next_page,temp_buff,keys,key_checksum,
		    level+1))
	goto err;
    }
    if (keypos >= endpos ||
	(key_length=(*keyinfo->get_key)(keyinfo,nod_flag,&keypos,key)) == 0)
      break;
    if ((*keys)++ &&
	(flag=_mi_key_cmp(keyinfo->seg,info->lastkey,key,key_length,
			  comp_flag, &not_used)) >=0)
    {
      DBUG_DUMP("old",(byte*) info->lastkey, info->lastkey_length);
      DBUG_DUMP("new",(byte*) key, key_length);

      if (comp_flag & SEARCH_FIND && flag == 0)
	print_error("Found dupplicated key at page %s",llstr(page,llbuff));
      else
	print_error("Key in wrong position at page %s",llstr(page,llbuff));
      goto err;
    }
    if (testflag & T_STATISTICS)
    {
      if (*keys == 1L)				/* first_key */
	unique_count[keyinfo->keysegs]++;
      else
      {
	uint diff;
	_mi_key_cmp(keyinfo->seg,info->lastkey,key,USE_WHOLE_KEY,SEARCH_FIND,
		    &diff);
	unique_count[diff-1]++;
      }
    }
    (*key_checksum)+= mi_byte_checksum(key,
				       key_length- info->s->rec_reflength);
    memcpy((char*) info->lastkey,(char*) key,key_length);
    info->lastkey_length=key_length;
    record= _mi_dpos(info,0,key+key_length);
    if (record >= info->s->state.data_file_length)
    {
#ifndef DBUG_OFF
      char llbuff3[22];
#endif
      print_error("Found key at page %s that points to record outside datafile",llstr(page,llbuff));
      DBUG_PRINT("test",("page: %s  record: %s  filelength: %s",
			 llstr(page,llbuff),llstr(record,llbuff2),
			 llstr(info->s->state.data_file_length,llbuff3)));
      DBUG_DUMP("key",(byte*) info->lastkey,key_length);
      goto err;
    }
    record_checksum+=record;
  }
  if (keypos != endpos)
  {
    print_error("Keyblock size at page %s is not correct.  Block length: %d  key length: %d",
                llstr(page,llbuff), used_length, (keypos - buff));
    goto err;
  }
  my_afree((byte*) temp_buff);
  DBUG_RETURN(0);
 err:
  my_afree((byte*) temp_buff);
  DBUG_RETURN(1);
} /* chk_index */


	/* Calculate a checksum of 1+2+3+4...N = N*(N+1)/2 without overflow */

static ha_checksum calc_checksum(ha_rows count)
{
  ulonglong sum,a,b;
  DBUG_ENTER("calc_checksum");

  sum=0;
  a=count; b=count+1;
  if (a & 1)
    b>>=1;
  else
    a>>=1;
  while (b)
  {
    if (b & 1)
      sum+=a;
    a<<=1; b>>=1;
  }
  DBUG_PRINT("exit",("sum: %lx",(ulong) sum));
  DBUG_RETURN((ha_checksum) sum);
} /* calc_checksum */


	/* Calc length of key in normal isam */

static uint isam_key_length(MI_INFO *info, register MI_KEYDEF *keyinfo)
{
  uint length;
  MI_KEYSEG *keyseg;
  DBUG_ENTER("isam_key_length");

  length= info->s->rec_reflength;
  for (keyseg=keyinfo->seg ; keyseg->type ; keyseg++)
    length+= keyseg->length;

  DBUG_PRINT("exit",("length: %d",length));
  DBUG_RETURN(length);
} /* key_length */


	/* Check that record-link is ok */

static int chk_data_link(MI_INFO *info,int extend)
{
  int	error,got_error,flag;
  uint	key,left_length,b_type,field;
  ha_rows records,del_blocks;
  my_off_t used,empty,pos,splits,start_recpos,
	   del_length,link_used,start_block;
  byte	*record,*to;
  char llbuff[22],llbuff2[22],llbuff3[22];
  ha_checksum intern_record_checksum;
  ha_checksum key_checksum[MI_MAX_POSSIBLE_KEY];
  my_bool static_row_size;
  MI_KEYDEF *keyinfo;
  MI_BLOCK_INFO block_info;
  DBUG_ENTER("chk_data_link");

  if (!(testflag & T_SILENT))
  {
    if (extend)
      puts("- check records and index references");
    else
      puts("- check record links");
  }

  if (!(record= (byte*) my_alloca(info->s->base.pack_reclength)))
  {
    print_error("Not Enough memory");
    DBUG_RETURN(-1);
  }
  records=used=link_used=splits=del_blocks=del_length=0;
  intern_record_checksum=crc=0;
  LINT_INIT(left_length);  LINT_INIT(start_recpos);  LINT_INIT(to);
  got_error=error=0;
  empty=pos=info->s->pack.header_length;

  /* Check how to calculate checksum of rows */
  static_row_size=1;
  if (info->s->data_file_type == COMPRESSED_RECORD)
  {
    for (field=0 ; field < info->s->base.fields ; field++)
    {
      if (info->s->rec[field].base_type == FIELD_BLOB ||
	  info->s->rec[field].base_type == FIELD_VARCHAR)
      {
	static_row_size=0;
	break;
      }
    }
  }

  bzero((char*) key_checksum,info->s->base.keys * sizeof(key_checksum[0]));
  while (pos < info->s->state.data_file_length)
  {
    switch (info->s->data_file_type) {
    case STATIC_RECORD:
      if (my_b_read(&read_cache,(byte*) record,info->s->base.pack_reclength))
	goto err;
      start_recpos=pos;
      pos+=info->s->base.pack_reclength;
      splits++;
      if (*record == '\0')
      {
	del_blocks++;
	del_length+=info->s->base.pack_reclength;
	continue;					/* Record removed */
      }
      crc+= mi_static_checksum(info,record);
      used+=info->s->base.pack_reclength;
      break;
    case DYNAMIC_RECORD:
      flag=block_info.second_read=0;
      block_info.next_filepos=pos;
      do
      {
	if (_mi_read_cache(&read_cache,(byte*) block_info.header,
			  (start_block=block_info.next_filepos),
			  sizeof(block_info.header),test(! flag) | 2))
	  goto err;
	if (start_block & (MI_DYN_ALIGN_SIZE-1))
	{
	  print_error("Wrong aligned block at %s",llstr(start_block,llbuff));
	  goto err2;
	}
	b_type=_mi_get_block_info(&block_info,-1,start_block);
	if (b_type & (BLOCK_DELETED | BLOCK_ERROR | BLOCK_SYNC_ERROR |
		      BLOCK_FATAL_ERROR))
	{
	  if (b_type & BLOCK_SYNC_ERROR)
	  {
	    if (flag)
	    {
	      print_error("Unexpected byte: %d at link: %s",
			  (int) block_info.header[0],
			  llstr(start_block,llbuff));
	      goto err2;
	    }
	    pos=block_info.filepos+block_info.block_len;
	    goto next;
	  }
	  if (b_type & BLOCK_DELETED)
	  {
	    if (block_info.block_len < info->s->base.min_block_length)
	    {
	      print_error("Deleted block with impossible length %lu at %s",
			  block_info.block_len,llstr(pos,llbuff));
	      goto err2;
	    }
	    if (block_info.next_filepos != HA_OFFSET_ERROR &&
		block_info.next_filepos >= info->s->state.data_file_length ||
		block_info.prev_filepos != HA_OFFSET_ERROR &&
		block_info.prev_filepos >= info->s->state.data_file_length)
	    {
	      print_error("Delete link points outside datafile at %s",
			  llstr(pos,llbuff));
	      goto err2;
	    }
	    del_blocks++;
	    del_length+=block_info.block_len;
	    pos=block_info.filepos+block_info.block_len;
	    splits++;
	    goto next;
	  }
	  print_error("Wrong bytesec: %d-%d-%d at linkstart: %s",
		      block_info.header[0],block_info.header[1],
		      block_info.header[2],
		      llstr(start_block,llbuff));
	  goto err2;
	}
	if (info->s->state.data_file_length < block_info.filepos+
	    block_info.block_len)
	{
	  print_error("Recordlink that points outside datafile at %s",
		      llstr(pos,llbuff));
	  got_error=1;
	  break;
	}
	splits++;
	if (!flag++)				/* First block */
	{
	  start_recpos=pos;
	  pos=block_info.filepos+block_info.block_len;
	  if (block_info.rec_len > (uint) info->s->base.max_pack_length)
	  {
	    print_error("Found too long record (%d) at %s",
			block_info.rec_len,
			llstr(start_recpos,llbuff));
	    got_error=1;
	    break;
	  }
	  if (info->s->base.blobs)
	  {
	    if (!(to=mi_fix_rec_buff_for_blob(info,block_info.rec_len)))
	    {
	      print_error("Not enough memory for blob at %s",
			  llstr(start_recpos,llbuff));
	      got_error=1;
	      break;
	    }
	  }
	  else
	    to= info->rec_buff;
	  left_length=block_info.rec_len;
	}
	if (left_length < block_info.data_len)
	{
	  print_error("Found too long record at %s",
		      llstr(start_recpos,llbuff));
	  got_error=1; break;
	}
	if (_mi_read_cache(&read_cache,(byte*) to,block_info.filepos,
			  (uint) block_info.data_len, test(flag == 1)))
	  goto err;
	to+=block_info.data_len;
	link_used+= block_info.filepos-start_block;
	used+= block_info.filepos - start_block + block_info.data_len;
	empty+=block_info.block_len-block_info.data_len;
	left_length-=block_info.data_len;
	if (left_length)
	{
	  if (b_type & BLOCK_LAST)
	  {
	    print_error("Record link to short for record at %s",
			llstr(start_recpos,llbuff));
	    got_error=1;
	    break;
	  }
	  if (info->s->state.data_file_length < block_info.next_filepos)
	  {
	    print_error("Found next-recordlink that points outside datafile at %s",
			llstr(block_info.filepos,llbuff));
	    got_error=1;
	    break;
	  }
	}
      } while (left_length);
      if (! got_error)
      {
	if (_mi_rec_unpack(info,record,info->rec_buff,block_info.rec_len) ==
	    MY_FILE_ERROR)
	{
	  print_error("Found wrong record at %s", llstr(start_recpos,llbuff));
	  got_error=1;
	}
	else
	{
	  info->checksum=mi_checksum(info,record);
	  if (testflag & (T_EXTEND | T_MEDIUM | T_VERBOSE))
	  {
	    if (_mi_rec_check(info,record))
	    {
	      print_error("Found wrong packed record at %s",
			  llstr(start_recpos,llbuff));
	      got_error=1;
	    }
	  }
	  if (!got_error)
	    crc+= info->checksum;
	}
      }
      else if (!flag)
	pos=block_info.filepos+block_info.block_len;
      break;
    case COMPRESSED_RECORD:
      if (_mi_read_cache(&read_cache,(byte*) block_info.header, pos,
			 info->s->pack.ref_length, 1))
	goto err;
      start_recpos=pos;
      splits++;
      VOID(_mi_pack_get_block_info(info,&block_info, -1, start_recpos));
      pos=block_info.filepos+block_info.rec_len;
      if (block_info.rec_len < (uint) info->s->min_pack_length ||
	  block_info.rec_len > (uint) info->s->max_pack_length)
      {
	print_error("Found block with wrong recordlength: %d at %s",
		    block_info.rec_len, llstr(start_recpos,llbuff));
	got_error=1;
	break;
      }
      if (_mi_read_cache(&read_cache,(byte*) info->rec_buff,
			block_info.filepos, block_info.rec_len,1))
	goto err;
      if (_mi_pack_rec_unpack(info,record,info->rec_buff,block_info.rec_len))
      {
	print_error("Found wrong record at %s", llstr(start_recpos,llbuff));
	got_error=1;
      }
      if (static_row_size)
	crc+= mi_static_checksum(info,record);
      else
	crc+= mi_checksum(info,record);
      link_used+= (block_info.filepos - start_recpos);
      used+= (pos-start_recpos);
    }
    if (! got_error)
    {
      intern_record_checksum+=start_recpos;
      records++;
      if (testflag & T_WRITE_LOOP && records % WRITE_COUNT == 0)
      {
	printf("%s\r", llstr(records,llbuff)); VOID(fflush(stdout));
      }

      /* Check if keys match the record */

	for (key=0,keyinfo= info->s->keyinfo; key < info->s->base.keys;
	     key++,keyinfo++)
	{
	  if ((((ulonglong) 1 << key) & info->s->state.key_map))
	  {
	    uint key_length=_mi_make_key(info,key,info->lastkey,record,
					 start_recpos);
	    if (extend)
	    {
	      if (_mi_search(info,keyinfo,info->lastkey,key_length,
			     SEARCH_SAME, info->s->state.key_root[key]))
	      {
		print_error("Record at: %10s  Can't find key for index: %2d",
			    llstr(start_recpos,llbuff),key+1);
		if (error++ > MAXERR || !(testflag & T_VERBOSE))
		  goto err2;
	      }
	    }
	    else
	      key_checksum[key]+=mi_byte_checksum(info->lastkey,key_length);
	  }
	}
    }
    else
    {
      got_error=0;
      if (error++ > MAXERR || !(testflag & T_VERBOSE))
	goto err2;
    }
  next:;				/* Next record */
  }
  if (testflag & T_WRITE_LOOP)
  {
    VOID(fputs("          \r",stdout)); VOID(fflush(stdout));
  }
  if (records != info->s->state.records)
  {
    print_error("Record-count is not ok; is %-10s   Should be: %s",
		llstr(records,llbuff), llstr(info->s->state.records,llbuff2));
    error=1;
  }
  else if (record_checksum != intern_record_checksum && info->s->state.key_map)
  {
    print_error("Keypointers and record positions doesn't match");
    error=1;
  }
  else if (crc != info->s->state.checksum &&
	   (info->s->options &
	    (HA_OPTION_CHECKSUM | HA_OPTION_COMPRESS_RECORD)))
  {
    print_warning("Record checksum is not the same as checksum stored in the index file\n");
    error=1;
  }
  else if (!extend)
  {
    for (key=0 ; key < info->s->base.keys;  key++)
    {
      if (key_checksum[key] != key_crc[key])
      {
	print_error("Checksum for key: %2d doesn't match checksum for records",
		    key+1);
	error=1;
      }
    }
  }

  if (used+empty+del_length != info->s->state.data_file_length)
  {
    print_warning("Found %s record-data and %s unused data and %s deleted-data",
		  llstr(used,llbuff),llstr(empty,llbuff2),
		  llstr(del_length,llbuff3));
    print_warning("Total %s, Should be: %s",
		  llstr((used+empty+del_length),llbuff),
		  llstr(info->s->state.data_file_length,llbuff2));
  }
  if (del_blocks != info->s->state.del)
  {
    print_warning("Found %10s deleted blocks       Should be: %s",
		  llstr(del_blocks,llbuff), llstr(info->s->state.del,llbuff2));
  }
  if (splits != info->s->state.split)
  {
    print_warning("Found %10s parts                Should be: %s parts",
		  llstr(splits,llbuff), llstr(info->s->state.split,llbuff2));
  }
  if (testflag & T_INFO)
  {
    if (warning_printed || error_printed)
      puts("");
    if (used != 0 && ! error_printed)
    {
      printf("Records:%18s    M.recordlength:%9lu   Packed:%14.0f%%\n",
	     llstr(records,llbuff), (used-link_used)/records,
	     (info->s->base.blobs ? 0 :
	      (ulonglong2double((ulonglong) info->s->base.reclength*records)-used)/
	      ulonglong2double((ulonglong) info->s->base.reclength*records)*100.0));
      printf("Recordspace used:%9.0f%%   Empty space:%12d%%  Blocks/Record: %6.2f\n",
	     (ulonglong2double(used-link_used)/ulonglong2double(used-link_used+empty)*100.0),
	     (!records ? 100 : (int) (ulonglong2double(del_length+empty)/used*100.0)),
	     ulonglong2double(splits - del_blocks) / records);
    }
    printf("Record blocks:%12s    Delete blocks:%10s\n",
	   llstr(splits-del_blocks,llbuff),llstr(del_blocks,llbuff2));
    printf("Record data:  %12s    Deleted data: %10s\n",
	   llstr(used-link_used,llbuff),llstr(del_length,llbuff2));
    printf("Lost space:   %12s    Linkdata:     %10s\n",
	   llstr(empty,llbuff),llstr(link_used,llbuff2));
  }
  my_afree((gptr) record);
  DBUG_RETURN (error);
 err:
  print_error("got error: %d when reading datafile",my_errno);
 err2:
  my_afree((gptr) record);
  DBUG_RETURN(1);
} /* chk_data_link */


	/* Recover old table by reading each record and writing all keys */
	/* Save new datafile-name in temp_filename */

static int rep(register MI_INFO *info, my_string name, int rep_quick)
{
  int error,got_error;
  uint i;
  ha_rows start_records,new_header_length;
  my_off_t del;
  File new_file;
  MYISAM_SHARE *share=info->s;
  char llbuff[22],llbuff2[22];
  DBUG_ENTER("rep");

  start_records=share->state.records;
  new_header_length=(testflag & T_UNPACK) ? 0L : share->pack.header_length;
  got_error=1;
  new_file= -1;
  if (!(testflag & T_SILENT))
  {
    printf("- recovering MYISAM-table '%s'\n",name);
    printf("Data records: %s\n", llstr(share->state.records,llbuff));
  }

  VOID(init_key_cache(use_buffers,NEAD_MEM));
  if (init_io_cache(&read_cache,info->dfile,(uint) read_buffer_length,
		   READ_CACHE,share->pack.header_length,1,MYF(MY_WME)))
    goto err;
  if (!rep_quick)
    if (init_io_cache(&info->rec_cache,-1,(uint) write_buffer_length,
		      WRITE_CACHE, new_header_length, 1,
		      MYF(MY_WME | MY_WAIT_IF_FULL)))
      goto err;
  info->opt_flag|=WRITE_CACHE_USED;
  sort_info.start_recpos=0;
  sort_info.buff=0; sort_info.buff_length=0;
  if (!(sort_info.record=(byte*) my_malloc((uint) share->base.pack_reclength,
					   MYF(0))))
  {
    print_error("Not Enough memory for extra record");
    goto err;
  }

  if (!rep_quick)
  {
    if ((new_file=my_create(fn_format(temp_filename,name,"",DATA_TMP_EXT,
				      2+4),
			    0,tmpfile_createflag,MYF(0))) < 0)
    {
      print_error("Can't create new tempfile: '%s'",temp_filename);
      goto err;
    }
    if (filecopy(new_file,info->dfile,0L,new_header_length,"datafile-header"))
      goto err;
    share->state.dellink= HA_OFFSET_ERROR;
    info->rec_cache.file=new_file;
    if (testflag & T_UNPACK)
    {
      share->options&= ~HA_OPTION_COMPRESS_RECORD;
      mi_int2store(share->state.header.options,share->options);
    }
  }
  sort_info.info=info;
  sort_info.pos=sort_info.max_pos=share->pack.header_length;
  sort_info.filepos=new_header_length;
  read_cache.end_of_file=sort_info.filelength=
    my_seek(info->dfile,0L,MY_SEEK_END,MYF(0));
  sort_info.dupp=0;
  sort_info.fix_datafile= (my_bool) (! rep_quick);
  sort_info.max_records=LONG_MAX;
  if ((sort_info.new_data_file_type=share->data_file_type) ==
      COMPRESSED_RECORD && testflag & T_UNPACK)
  {
    if (share->options & HA_OPTION_PACK_RECORD)
      sort_info.new_data_file_type = DYNAMIC_RECORD;
    else
      sort_info.new_data_file_type = STATIC_RECORD;
  }

  del=share->state.del;
  share->state.records=share->state.del=share->state.empty=
    share->state.split=0;
  if (sort_info.new_data_file_type != COMPRESSED_RECORD)
    share->state.checksum=0;
  info->update= (short) HA_STATE_CHANGED;
  for (i=0 ; i < info->s->base.keys ; i++)
    share->state.key_root[i]= HA_OFFSET_ERROR;
  for (i=0 ; i < share->state.header.max_block_size ; i++)
    share->state.key_del[i]=  HA_OFFSET_ERROR;

  share->state.key_file_length=share->base.keystart;

  lock_memory();				/* Everything is alloced */
  while (!(error=sort_get_next_record()))
  {
    if (writekeys(info,(byte*) sort_info.record,sort_info.filepos))
    {
      if (my_errno != HA_ERR_FOUND_DUPP_KEY) goto err;
      DBUG_DUMP("record",(byte*) sort_info.record,share->base.pack_reclength);
      print_info("Dupplicate key %2d for record at %10s against new record at %10s",
		 info->errkey+1,
		 llstr(sort_info.start_recpos,llbuff),
		 llstr(info->int_pos,llbuff2));
      if (testflag & T_VERBOSE)
      {
	VOID(_mi_make_key(info,(uint) info->errkey,info->lastkey,
			  sort_info.record,0L));
	_mi_print_key(stdout,share->keyinfo[info->errkey].seg,info->lastkey);
      }
      sort_info.dupp++;
      if (rep_quick == 1)
      {
	error_printed=1;
	goto err;
      }
      continue;
    }
    if (sort_write_record())
      goto err;
  }
  if (error > 0 || write_data_suffix(info) ||
      flush_io_cache(&info->rec_cache) || read_cache.error < 0)
    goto err;

  if (testflag & T_WRITE_LOOP)
  {
    VOID(fputs("          \r",stdout)); VOID(fflush(stdout));
  }
  if (my_chsize(share->kfile,share->state.key_file_length,MYF(0)))
  {
    print_warning("Can't change size of indexfile, error: %d",my_errno);
    goto err;
  }

  if (rep_quick && del+sort_info.dupp != share->state.del)
  {
    print_error("Couldn't fix table with quick recovery: Found wrong number of deleted records");
    print_error("Run recovery again without -q");
    got_error=1;
    goto err;
  }

  if (!rep_quick)
  {
    info->dfile=new_file;
    share->state.data_file_length=sort_info.filepos;
    share->state.split=share->state.records;		/* Only hole records */
    out_flag|=O_NEW_DATA;				/* Data in new file */
    share->state.version=(ulong) time((time_t*) 0);	/* Force reopen */
  }
  else
    share->state.data_file_length=sort_info.max_pos;

  if (!(testflag & T_SILENT))
  {
    if (start_records != share->state.records)
      printf("Data records: %s\n", llstr(share->state.records,llbuff));
    if (sort_info.dupp)
      print_warning("%s records have been removed",
		    llstr(sort_info.dupp,llbuff));
  }

  got_error=0;
err:
  if (got_error)
  {
    if (! error_printed)
      print_error("%d for record at pos %s",my_errno,
		  llstr(sort_info.start_recpos,llbuff));
    if (new_file >= 0)
    {
      VOID(my_close(new_file,MYF(0)));
      VOID(my_delete(temp_filename,MYF(MY_WME)));
    }
  }
  if (sort_info.record)
  {
    my_afree(sort_info.record);
  }
  my_free(sort_info.buff,MYF(MY_ALLOW_ZERO_PTR));
  VOID(end_io_cache(&read_cache));
  info->opt_flag&= ~(READ_CACHE_USED | WRITE_CACHE_USED);
  VOID(end_io_cache(&info->rec_cache));
  got_error|=flush_blocks(share->kfile);
  if (!got_error && testflag & T_UNPACK)
  {
    share->state.header.options[0]&= (uchar) ~HA_OPTION_COMPRESS_RECORD;
    share->pack.header_length=0;
    share->data_file_type=sort_info.new_data_file_type;
  }
  DBUG_RETURN(got_error);
} /* rep */


	/* Uppdaterar nyckelfilen i samband med reparation */

static int writekeys(register MI_INFO *info,byte *buff,my_off_t filepos)
{
  register uint i;
  uchar *key;
  DBUG_ENTER("writekeys");

  key=info->lastkey+info->s->base.max_key_length;
  for (i=0 ; i < info->s->base.keys ; i++)
  {
    if (((ulonglong) 1 << i) & info->s->state.key_map)
    {
      uint key_length=_mi_make_key(info,i,key,buff,filepos);
      if (_mi_ck_write(info,i,key,key_length)) goto err;
    }
  }
  DBUG_RETURN(0);

 err:
  if (my_errno == HA_ERR_FOUND_DUPP_KEY)
  {
    info->errkey=(int) i;			/* This key was found */
    while ( i-- > 0 )
    {
      if (((ulonglong) 1 << i) & info->s->state.key_map)
      {
	uint key_length=_mi_make_key(info,i,key,buff,filepos);
	if (_mi_ck_delete(info,i,key,key_length)) break;
      }
    }
  }
  DBUG_PRINT("error",("errno: %d",my_errno));
  DBUG_RETURN(-1);
} /* writekeys */


	 /* Write info about table */

static void descript(register MI_INFO *info, my_string name)
{
  uint key,keyseg_nr,field,start,len;
  reg3 MI_KEYDEF *keyinfo;
  reg2 MI_KEYSEG *keyseg;
  reg4 char *text;
  char buff[40],length[10],*pos,*end;
  enum en_fieldtype type;
  MYISAM_SHARE *share=info->s;
  char llbuff[22],llbuff2[22];
  DBUG_ENTER("describe");

  printf("\nMYISAM file:         %s\n",name);
  fputs("Record format:       ",stdout);
  if (share->options & HA_OPTION_COMPRESS_RECORD)
    puts("Compressed");
  else if (share->options & HA_OPTION_PACK_RECORD)
    puts("Packed");
  else
    puts("Fixed length");

  if (testflag & T_VERBOSE)
  {
    printf("File-version:        %d\n",
	   (int) share->state.header.file_version[3]);
    if (share->state.create_time)
    {
      get_date(buff,1,share->state.create_time);
      printf("Creation time:       %s\n",buff);
    }
    if (share->state.check_time)
    {
      get_date(buff,1,share->state.check_time);
      printf("Recover time:        %s\n",buff);
    }
    printf("Status:              %s\n",
	   share->state.open_count ? "open" :
	   share->state.changed & 2 ? "crashed" :
	   share->state.changed ? "changed" :"checked");
    if (share->base.auto_key)
    {
      printf("Auto increment key:  %13d  Last value:         %13s\n",
	     share->base.auto_key,
	     llstr(share->state.auto_increment,llbuff));
    }
    if (share->options & HA_OPTION_CHECKSUM)
      printf("Using checksums\n");
  }
  printf("Data records:        %13s  Deleted blocks:     %13s\n",
	 llstr(share->state.records,llbuff),llstr(share->state.del,llbuff2));
  if (testflag & T_SILENT)
    DBUG_VOID_RETURN;				/* This is enough */

  if (testflag & T_VERBOSE)
  {
#ifdef USE_RELOC
    printf("Init-relocation:     %13s\n",llstr(share->base.reloc,llbuff));
#endif
    printf("Datafile parts:      %13s  Deleted data:       %13s\n",
	   llstr(share->state.split,llbuff),
	   llstr(share->state.empty,llbuff2));
    printf("Datafile pointer (bytes):%9d  Keyfile pointer (bytes):%9d\n",
	   share->rec_reflength,share->base.key_reflength);
    if (info->s->base.reloc == 1L && info->s->base.records == 1L)
      puts("This is a one-record table");
    else
    {
      if (share->base.max_data_file_length != HA_OFFSET_ERROR ||
	  share->base.max_key_file_length != HA_OFFSET_ERROR)
	printf("Max datafile length: %13s  Max keyfile length: %13s\n",
	       llstr(share->base.max_data_file_length-1,llbuff),
	       llstr(share->base.max_key_file_length-1,llbuff2));
    }
  }

  printf("Recordlength:        %13d\n",(int) share->base.pack_reclength);
  if (share->state.key_map != (((ulonglong) 1 << share->base.keys) -1))
  {
    char buff[33];
    longlong2str(share->state.key_map,buff,2);
    printf("Using only keys '%s' of %d possibly keys\n",
	   buff, share->base.keys);
  }
  puts("\ntable description:");
  printf("Key Start Len Index   Type");
  if (testflag & T_VERBOSE)
    printf("                     Rec/key         Root  Blocksize");
  VOID(putchar('\n'));

  for (key=keyseg_nr=0, keyinfo= &share->keyinfo[0] ;
       key < share->base.keys;
       key++,keyinfo++)
  {
    char buff[22];
    keyseg=keyinfo->seg;
    if (keyinfo->flag & HA_NOSAME) text="unique ";
    else text="multip.";

    pos=buff;
    if (keyseg->flag & HA_REVERSE_SORT)
      *pos++ = '-';
    pos=strmov(pos,type_names[keyseg->type]);
    *pos++ = ' ';
    *pos=0;
    if (keyinfo->flag & HA_PACK_KEY)
      pos=strmov(pos,prefix_packed_txt);
    if (keyinfo->flag & HA_BINARY_PACK_KEY)
      pos=strmov(pos,bin_packed_txt);
    if (keyseg->flag & HA_SPACE_PACK)
      pos=strmov(pos,diff_txt);
    if (keyseg->flag & HA_BLOB_PART)
      pos=strmov(pos,blob_txt);
    if (keyseg->flag & HA_NULL_PART)
      pos=strmov(pos,null_txt);
    *pos=0;

    printf("%-4d%-6d%-3d %-8s%-21s",
	   key+1,keyseg->start+1,keyseg->length,text,buff);
    if (share->state.key_root[key] != HA_OFFSET_ERROR)
      llstr(share->state.key_root[key],buff);
    else
      buff[0]=0;
    if (testflag & T_VERBOSE)
      printf("%11lu %12s %10d",
	     share->state.rec_per_key_part[keyseg_nr++],
	     buff,keyinfo->block_length);
    VOID(putchar('\n'));
    while ((++keyseg)->type != HA_KEYTYPE_END)
    {
      pos=buff;
      if (keyseg->flag & HA_REVERSE_SORT)
	*pos++ = '-';
      pos=strmov(pos,type_names[keyseg->type]);
      *pos++= ' ';
      if (keyseg->flag & HA_SPACE_PACK)
	pos=strmov(pos,diff_txt);
      if (keyseg->flag & HA_BLOB_PART)
	pos=strmov(pos,blob_txt);
      if (keyseg->flag & HA_NULL_PART)
	pos=strmov(pos,null_txt);
      *pos=0;
      printf("    %-6d%-3d         %-21s",
	     keyseg->start+1,keyseg->length,buff);
      if (testflag & T_VERBOSE)
	printf("%11lu", share->state.rec_per_key_part[keyseg_nr++]);
      VOID(putchar('\n'));
    }
    keyseg++;
  }
  if (share->state.header.uniques)
  {
    MI_UNIQUEDEF *uniqueinfo;
    puts("\nUnique  Key  Start  Len  Nullpos  Nullbit  Type");
    for (key=0,uniqueinfo= &share->uniqueinfo[0] ;
	 key < share->state.header.uniques; key++, uniqueinfo++)
    {
      my_bool new_row=0;
      char null_bit[8],null_pos[8];
      printf("%-8d%-5d",key+1,uniqueinfo->key+1);
      for ( ; keyseg->type != HA_KEYTYPE_END ; keyseg++)
      {
	if (new_row)
	  fputs("             ",stdout);
	null_bit[0]=null_pos[0]=0;
	if (keyseg->null_bit)
	{
	  sprintf(null_bit,"%d",keyseg->null_bit);
	  sprintf(null_pos,"%d",keyseg->null_pos+1);
	}
	printf("%-7d%-5d%-9s%-10s%-30s\n",
	       keyseg->start+1,keyseg->length,
	       null_pos,null_bit,
	       type_names[keyseg->type]);
	new_row=1;
      }
    }
  }
  if (verbose > 1)
  {
    char null_bit[8],null_pos[8];
    printf("\nField Start Length Nullpos Nullbit Type");
    if (share->options & HA_OPTION_COMPRESS_RECORD)
      printf("                         Huff tree  Bits");
    VOID(putchar('\n'));
    start=1;
    for (field=0 ; field < share->base.fields ; field++)
    {
      if (share->options & HA_OPTION_COMPRESS_RECORD)
	type=share->rec[field].base_type;
      else
	type=(enum en_fieldtype) share->rec[field].type;
      end=strmov(buff,field_pack[type]);
      if (share->options & HA_OPTION_COMPRESS_RECORD)
      {
	if (share->rec[field].pack_type & PACK_TYPE_SELECTED)
	  end=strmov(end,", not_always");
	if (share->rec[field].pack_type & PACK_TYPE_SPACE_FIELDS)
	  end=strmov(end,", no empty");
	if (share->rec[field].pack_type & PACK_TYPE_ZERO_FILL)
	{
	  sprintf(end,", zerofill(%d)",share->rec[field].space_length_bits);
	  end=strend(end);
	}
      }
      if (buff[0] == ',')
	strmov(buff,buff+2);
      len=(uint) (int2str((long) share->rec[field].length,length,10) -
		  length);
      null_bit[0]=null_pos[0]=0;
      if (share->rec[field].null_bit)
      {
	sprintf(null_bit,"%d",share->rec[field].null_bit);
	sprintf(null_pos,"%d",share->rec[field].null_pos+1);
      }
      printf("%-6d%-6d%-7s%-8s%-8s%-35s",field+1,start,length,
	     null_pos, null_bit, buff);
      if (share->options & HA_OPTION_COMPRESS_RECORD)
      {
	if (share->rec[field].huff_tree)
	  printf("%3d    %2d",
		 (uint) (share->rec[field].huff_tree-share->decode_trees)+1,
		 share->rec[field].huff_tree->quick_table_bits);
      }
      VOID(putchar('\n'));
      start+=share->rec[field].length;
    }
  }
  DBUG_VOID_RETURN;
} /* describe */


	/* Change all key-pointers that points to a records */

static int movepoint(register MI_INFO *info, byte *record, my_off_t oldpos,
		     my_off_t newpos, uint prot_key)
{
  register uint i;
  uchar *key;
  uint key_length;
  DBUG_ENTER("movepoint");

  key=info->lastkey+info->s->base.max_key_length;
  for (i=0 ; i < info->s->base.keys; i++)
  {
    if (i != prot_key && (((ulonglong) 1 << i) & info->s->state.key_map))
    {
      key_length=_mi_make_key(info,i,key,record,oldpos);
      if (info->s->keyinfo[i].flag & HA_NOSAME)
      {					/* Change pointer direct */
	uint nod_flag;
	MI_KEYDEF *keyinfo;
	keyinfo=info->s->keyinfo+i;
	if (_mi_search(info,keyinfo,key,USE_WHOLE_KEY,
		       (uint) (SEARCH_SAME | SEARCH_SAVE_BUFF),
		       info->s->state.key_root[i]))
	  DBUG_RETURN(-1);
	nod_flag=mi_test_if_nod(info->buff);
	_mi_dpointer(info,info->int_keypos-nod_flag-
		     info->s->rec_reflength,newpos);
	if (_mi_write_keypage(info,keyinfo,info->int_pos,info->buff))
	  DBUG_RETURN(-1);
      }
      else
      {					/* Change old key to new */
	if (_mi_ck_delete(info,i,key,key_length))
	  DBUG_RETURN(-1);
	key_length=_mi_make_key(info,i,key,record,newpos);
	if (_mi_ck_write(info,i,key,key_length))
	  DBUG_RETURN(-1);
      }
    }
  }
  DBUG_RETURN(0);
} /* movepoint */


	/* Tell system that we want all memory for our cache */

static void lock_memory(void)
{
#ifdef SUN_OS				/* Key-cacheing thrases on sun 4.1 */
  int success;

  success = mlockall(MCL_CURRENT);	/* or plock(DATLOCK); */

  if (geteuid() == 0 && success != 0)
    print_warning("Failed to lock memory. errno %d",my_errno);
#endif
} /* lock_memory */


	/* Flush all changed blocks to disk */

static int flush_blocks(File file)
{
  if (flush_key_blocks(file,FLUSH_RELEASE))
  {
    print_error("%d when trying to write bufferts",my_errno);
    return(1);
  }
  end_key_cache();
  return 0;
} /* flush_blocks */


	/* Sort records according to one key */

static int sort_records(register MI_INFO *info, my_string name, uint sort_key,
			int write_info)
{
  int got_error;
  uint key;
  MI_KEYDEF *keyinfo;
  File new_file;
  uchar *temp_buff;
  ha_rows old_record_count;
  MYISAM_SHARE *share=info->s;
  char llbuff[22],llbuff2[22];
  DBUG_ENTER("sort_records");

  keyinfo= &share->keyinfo[sort_key];
  got_error=1;
  temp_buff=0; sort_info.record=0;
  new_file= -1;

  if (!(((ulonglong) 1 << sort_key) & share->state.key_map))
  {
    print_error("Can't sort table '%s' on key %d;  No such key",
		name,sort_key+1);
    error_printed=0;
    DBUG_RETURN(-1);
  }
  if (!(testflag & T_SILENT))
  {
    printf("- Sorting records for MYISAM-table '%s'\n",name);
    if (write_info)
      printf("Data records: %9s   Deleted: %9s\n",
	     llstr(share->state.records,llbuff),
	     llstr(share->state.del,llbuff2));
  }
  if (share->state.key_root[sort_key] == HA_OFFSET_ERROR)
    DBUG_RETURN(0);				/* Nothing to do */

  init_key_cache(use_buffers,NEAD_MEM);
  if (init_io_cache(&info->rec_cache,-1,(uint) write_buffer_length,
		   WRITE_CACHE,share->pack.header_length,1,
		   MYF(MY_WME | MY_WAIT_IF_FULL)))
    goto err;
  info->opt_flag|=WRITE_CACHE_USED;

  if (!(temp_buff=(uchar*) my_alloca((uint) keyinfo->block_length)))
  {
    print_error("Not enough memory for key block");
    goto err;
  }
  if (!(sort_info.record=(byte*) my_malloc((uint) share->base.pack_reclength,
					   MYF(0))))
  {
    print_error("Not enough memory for record");
    goto err;
  }
  if ((new_file=my_create(fn_format(temp_filename,name,"",DATA_TMP_EXT,2+4),
			  0,tmpfile_createflag,MYF(0))) <= 0)
  {
    print_error("Can't create new tempfile: '%s'",temp_filename);
    goto err;
  }
  if (filecopy(new_file,info->dfile,0L,share->pack.header_length,
	       "datafile-header"))
    goto err;
  info->rec_cache.file=new_file;		/* Use this file for cacheing*/

  lock_memory();
  for (key=0 ; key < share->base.keys ; key++)
    share->keyinfo[key].flag|= HA_SORT_ALLOWS_SAME;

  if (my_pread(share->kfile,(byte*) temp_buff,
	       (uint) keyinfo->block_length,
	       share->state.key_root[sort_key],
	       MYF(MY_NABP+MY_WME)))
  {
    print_error("Can't read indexpage from filepos: %s",
		(ulong) share->state.key_root[sort_key]);
    goto err;
  }

  /* Setup param for sort_write_record */
  sort_info.info=info;
  sort_info.new_data_file_type=share->data_file_type;
  sort_info.fix_datafile=1;
  sort_info.filepos=share->pack.header_length;
  old_record_count=share->state.records;
  share->state.records=0;
  if (sort_info.new_data_file_type != COMPRESSED_RECORD)
    share->state.checksum=0;

  if (sort_record_index(info,keyinfo,share->state.key_root[sort_key],temp_buff,
			sort_key,new_file) ||
      write_data_suffix(info) ||
      flush_io_cache(&info->rec_cache))
    goto err;

  if (share->state.records != old_record_count)
  {
    print_error("found %s of %s records",
		llstr(share->state.records,llbuff),
		llstr(old_record_count,llbuff2));
    goto err;
  }

	/* Put same locks as old file */
  if (lock_file(new_file,0L,F_WRLCK,"tempfile",temp_filename))
    goto err;
  VOID(lock_file(info->dfile,0L,F_UNLCK,"datafile of",name));
  VOID(my_close(info->dfile,MYF(MY_WME)));
  out_flag|=O_NEW_DATA;				/* Data in new file */

  info->dfile=new_file;				/* Use new indexfile */
  share->state.del=share->state.empty=0;
  share->state.dellink= HA_OFFSET_ERROR;
  share->state.data_file_length=sort_info.filepos;
  share->state.split=share->state.records;	/* Only hole records */
  share->state.version=(ulong) time((time_t*) 0);

  info->update= (short) HA_STATE_CHANGED;

  if (testflag & T_WRITE_LOOP)
  {
    VOID(fputs("          \r",stdout)); VOID(fflush(stdout));
  }
  got_error=0;

err:
  if (got_error && new_file >= 0)
  {
    VOID(my_close(new_file,MYF(MY_WME)));
    VOID(my_delete(temp_filename,MYF(MY_WME)));
  }
  if (temp_buff)
  {
    my_afree((gptr) temp_buff);
  }
  my_free(sort_info.record,MYF(MY_ALLOW_ZERO_PTR));
  info->opt_flag&= ~(READ_CACHE_USED | WRITE_CACHE_USED);
  VOID(end_io_cache(&info->rec_cache));
  share->state.sortkey=sort_key;
  DBUG_RETURN(flush_blocks(share->kfile) | got_error);
} /* sort_records */


	 /* Sort records recursive using one index */

static int sort_record_index(MI_INFO *info, MI_KEYDEF *keyinfo, my_off_t page,
			     uchar *buff, uint sort_key, File new_file)
{
  uint	nod_flag,used_length,key_length;
  uchar *temp_buff,*keypos,*endpos;
  my_off_t next_page,rec_pos;
  uchar lastkey[MI_MAX_KEY_BUFF];
  char llbuff[22];
  DBUG_ENTER("sort_record_index");

  nod_flag=mi_test_if_nod(buff);
  temp_buff=0;

  if (nod_flag)
  {
    if (!(temp_buff=(uchar*) my_alloca((uint) keyinfo->block_length)))
    {
      print_error("Not Enough memory");
      DBUG_RETURN(-1);
    }
  }
  used_length=mi_getint(buff);
  keypos=buff+2+nod_flag;
  endpos=buff+used_length;
  for ( ;; )
  {
    _sanity(__FILE__,__LINE__);
    if (nod_flag)
    {
      next_page=_mi_kpos(nod_flag,keypos);
      if (my_pread(info->s->kfile,(byte*) temp_buff,
		  (uint) keyinfo->block_length, next_page,
		   MYF(MY_NABP+MY_WME)))
      {
	print_error("Can't read keys from filepos: %s",
		    llstr(next_page,llbuff));
	goto err;
      }
      if (sort_record_index(info,keyinfo,next_page,temp_buff,sort_key,
			    new_file))
	goto err;
    }
    _sanity(__FILE__,__LINE__);
    if (keypos >= endpos ||
	(key_length=(*keyinfo->get_key)(keyinfo,nod_flag,&keypos,lastkey))
	== 0)
      break;
    rec_pos= _mi_dpos(info,0,lastkey+key_length);

    if ((*info->s->read_rnd)(info,sort_info.record,rec_pos,0))
    {
      print_error("%d when reading datafile",my_errno);
      goto err;
    }
    if (rec_pos != sort_info.filepos)
    {
      _mi_dpointer(info,keypos-nod_flag-info->s->rec_reflength,
		   sort_info.filepos);
      if (movepoint(info,sort_info.record,rec_pos,sort_info.filepos,
		    sort_key))
      {
	print_error("%d when updating key-pointers",my_errno);
	goto err;
      }
    }
    if (sort_write_record())
      goto err;
  }
  bzero((byte*) buff+used_length,keyinfo->block_length-used_length);
  if (my_pwrite(info->s->kfile,(byte*) buff,(uint) keyinfo->block_length,
		page,MYF_RW))
  {
    print_error("%d when updating keyblock",my_errno);
    goto err;
  }
  if (temp_buff)
    my_afree((gptr) temp_buff);
  DBUG_RETURN(0);
err:
  if (temp_buff)
    my_afree((gptr) temp_buff);
  DBUG_RETURN(1);
} /* sort_record_index */


	/* Sort index for more efficent reads */

static int sort_index(register MI_INFO *info, my_string name)
{
  reg2 uint key;
  reg1 MI_KEYDEF *keyinfo;
  File new_file;
  my_off_t index_pos[MI_MAX_POSSIBLE_KEY];
  char llbuff[22];
  DBUG_ENTER("sort_index");

  if (!(testflag & T_SILENT))
    printf("- Sorting index for MYISAM-table '%s'\n",name);

  if ((new_file=my_create(fn_format(temp_filename,name,"",INDEX_TMP_EXT,2+4),
			  0,tmpfile_createflag,MYF(0))) <= 0)
  {
    print_error("Can't create new tempfile: '%s'",temp_filename);
    DBUG_RETURN(-1);
  }
  if (filecopy(new_file,info->s->kfile,0L,(ulong) info->s->base.keystart,
	       "headerblock"))
    goto err;

  new_file_pos=info->s->base.keystart;
  for (key= 0,keyinfo= &info->s->keyinfo[0]; key < info->s->base.keys ;
       key++,keyinfo++)
  {
    if (!(((ulonglong) 1 << key) & info->s->state.key_map))
      continue;

    if (info->s->state.key_root[key] != HA_OFFSET_ERROR)
    {
      index_pos[key]=new_file_pos;		/* Write first block here */
      if (!_mi_fetch_keypage(info,keyinfo,info->s->state.key_root[key],
			     info->buff,0))
      {
	print_error("Can't read indexpage from filepos: %s",
		    llstr(info->s->state.key_root[key],llbuff));
	goto err;
      }
      if (sort_one_index(info,keyinfo,info->buff,new_file))
	goto err;
    }
    else
      index_pos[key]= HA_OFFSET_ERROR;		/* No blocks */
  }

	/* Put same locks as old file */
  if (lock_file(new_file,0L,F_WRLCK,"tempfile",temp_filename))
    goto err;
  info->s->state.version=(ulong) time((time_t*) 0);
  VOID(_mi_writeinfo(info,WRITEINFO_UPDATE_KEYFILE));
  VOID(my_close(info->s->kfile,MYF(MY_WME)));
  out_flag|=O_NEW_INDEX;			/* Data in new file */

  info->s->kfile=new_file;
  info->s->state.key_file_length=new_file_pos;
  info->update= (short) HA_STATE_CHANGED;
  for (key=0 ; key < info->s->base.keys ; key++)
    info->s->state.key_root[key]=index_pos[key];
  for (key=0 ; key < info->s->state.header.max_block_size ; key++)
    info->s->state.key_del[key]=  HA_OFFSET_ERROR;

  DBUG_RETURN(0);

err:
  VOID(my_close(new_file,MYF(MY_WME)));
  VOID(my_delete(temp_filename,MYF(MY_WME)));
  DBUG_RETURN(-1);
} /* sort_index */


	 /* Sort records recursive using one index */

static int sort_one_index(MI_INFO *info, MI_KEYDEF *keyinfo, uchar *buff,
			  File new_file)
{
  uint length,nod_flag,used_length;
  uchar *temp_buff,*keypos,*endpos;
  my_off_t new_page_pos,next_page;
  char llbuff[22];
  DBUG_ENTER("sort_one_index");

  temp_buff=0;
  new_page_pos=new_file_pos;
  new_file_pos+=keyinfo->block_length;

  if ((nod_flag=mi_test_if_nod(buff)))
  {
    if (!(temp_buff=(uchar*) my_alloca((uint) keyinfo->block_length)))
    {
      print_error("Not Enough memory");
      DBUG_RETURN(-1);
    }

    used_length=mi_getint(buff);
    keypos=buff+2+nod_flag;
    endpos=buff+used_length;
    for ( ;; )
    {
      if (nod_flag)
      {
	next_page=_mi_kpos(nod_flag,keypos);
	_mi_kpointer(info,keypos-nod_flag,new_file_pos); /* Save new pos */
	if (!_mi_fetch_keypage(info,keyinfo,next_page,temp_buff,0))
	{
	  print_error("Can't read keys from filepos: %s",
		      llstr(next_page,llbuff));
	  goto err;
	}
	if (sort_one_index(info,keyinfo,temp_buff,new_file))
	  goto err;
      }
      if (keypos >= endpos ||
	  ((*keyinfo->get_key)(keyinfo,nod_flag,&keypos,info->lastkey)) == 0)
	break;
    }
    my_afree((gptr) temp_buff);
  }

	/* Fill block with zero and write it to new file */

  length=mi_getint(buff);
  bzero((byte*) buff+length,keyinfo->block_length-length);
  if (my_pwrite(new_file,(byte*) buff,(uint) keyinfo->block_length,
		new_page_pos,MYF(MY_NABP | MY_WAIT_IF_FULL)))
  {
    print_error("Can't write indexblock, error: %d",my_errno);
    goto err;
  }
  DBUG_RETURN(0);
err:
  if (temp_buff)
    my_afree((gptr) temp_buff);
  DBUG_RETURN(1);
} /* sort_one_index */


	/* Change to use new file */
	/* Copy stats from old file to new file, deletes orginal and */
	/* changes new file name to old file name */

static int change_to_newfile(my_string filename, my_string old_ext,
			     my_string new_ext)
{
  char old_filename[FN_REFLEN],new_filename[FN_REFLEN];

  return my_redel(fn_format(old_filename,filename,"",old_ext,2+4),
		  fn_format(new_filename,filename,"",new_ext,2+4),
		  MYF(MY_WME+MY_LINK_WARNING));
} /* change_to_newfile */


	/* Locks a whole file */
	/* Gives an error-message if file can't be locked */

static int lock_file(File file, my_off_t start, int lock_type,
		     my_string filetype, my_string filename)
{
#ifndef NO_LOCKING
  if (my_lock(file,lock_type,start,F_TO_EOF,
	      testflag & T_WAIT_FOREVER ? MYF(MY_SEEK_NOT_DONE) :
	      MYF(MY_SEEK_NOT_DONE |  MY_DONT_WAIT)))
  {
    print_error(" %d when locking %s '%s'",my_errno,filetype,filename);
    error_printed=2;		/* Don't give that data is crashed */
    return 1;
  }
#endif
  return 0;
} /* lock_file */


	/* Copy a block between two files */

static int filecopy(File to,File from,my_off_t start,my_off_t length,
		    my_string type)
{
  char tmp_buff[IO_SIZE],*buff;
  ulong buff_length;
  DBUG_ENTER("filecopy");

  buff_length=min(write_buffer_length,length);
  if (!(buff=my_malloc(buff_length,MYF(0))))
  {
    buff=tmp_buff; buff_length=IO_SIZE;
  }

  VOID(my_seek(from,start,MY_SEEK_SET,MYF(0)));
  while (length > buff_length)
  {
    if (my_read(from,(byte*) buff,buff_length,MYF(MY_NABP)) ||
	my_write(to,(byte*) buff,buff_length,MYF(MY_NABP | MY_WAIT_IF_FULL)))
      goto err;
    length-= buff_length;
  }
  if (my_read(from,(byte*) buff,(uint) length,MYF(MY_NABP)) ||
      my_write(to,(byte*) buff,(uint) length,MYF(MY_NABP | MY_WAIT_IF_FULL)))
    goto err;
  if (buff != tmp_buff)
    my_free(buff,MYF(0));
  DBUG_RETURN(0);
err:
  if (buff != tmp_buff)
    my_free(buff,MYF(0));
  print_error("Can't copy %s to tempfile, error %d",type,my_errno);
  DBUG_RETURN(1);
}

	/* Fix table using sorting */
	/* saves new table in temp_filename */

static int rep_by_sort(register MI_INFO *info, my_string name,
		       int rep_quick)
{
  int got_error;
  uint i;
  ulong length;
  ha_rows start_records;
  my_off_t new_header_length,del;
  File new_file;
  MI_SORT_PARAM sort_param;
  MYISAM_SHARE *share=info->s;
  ulong   *rec_per_key_part;
  char llbuff[22];
  DBUG_ENTER("rep_by_sort");

  start_records=share->state.records;
  got_error=1;
  new_file= -1;
  sort_info.key_block=0; sort_info.record=0;
  new_header_length=(testflag & T_UNPACK) ? 0 : share->pack.header_length;
  if (!(testflag & T_SILENT))
  {
    printf("- recovering MYISAM-table '%s'\n",name);
    printf("Data records: %s\n", llstr(start_records,llbuff));
  }
  if (!(sort_info.key_block=alloc_key_blocks((uint) sort_key_blocks,
					     share->base.max_key_block_length))
      || init_io_cache(&read_cache,info->dfile,(uint) read_buffer_length,
		      READ_CACHE,share->pack.header_length,1,MYF(MY_WME)) ||
      (! rep_quick &&
       init_io_cache(&info->rec_cache,info->dfile,(uint) write_buffer_length,
		    WRITE_CACHE,new_header_length,1,
		    MYF(MY_WME | MY_WAIT_IF_FULL))))
    goto err;
  sort_info.key_block_end=sort_info.key_block+sort_key_blocks;
  info->opt_flag|=WRITE_CACHE_USED;
  info->rec_cache.file=info->dfile;		/* for sort_delete_record */

  if (!(sort_info.record=(byte*) my_malloc((uint) share->base.pack_reclength,
					   MYF(0))))
  {
    print_error("Not enough memory for extra record");
    goto err;
  }
  if (!rep_quick)
  {
    if ((new_file=my_create(fn_format(temp_filename,name,"",DATA_TMP_EXT,
				      2+4),
			    0,tmpfile_createflag,MYF(0))) < 0)
    {
      print_error("Can't create new tempfile: '%s'",temp_filename);
      goto err;
    }
    if (filecopy(new_file,info->dfile,0L,new_header_length,"datafile-header"))
      goto err;
    if (testflag & T_UNPACK)
    {
      share->options&= ~HA_OPTION_COMPRESS_RECORD;
      mi_int2store(share->state.header.options,share->options);
    }
    share->state.dellink= HA_OFFSET_ERROR;
    info->rec_cache.file=new_file;
  }

  info->update= (short) HA_STATE_CHANGED;
  for (i=0 ; i < share->base.keys ; i++)
    share->state.key_root[i]= HA_OFFSET_ERROR;
  for (i=0 ; i < share->state.header.max_block_size ; i++)
    share->state.key_del[i]=  HA_OFFSET_ERROR;

  share->state.key_file_length=share->base.keystart;

  sort_info.info=info;
  if ((sort_info.new_data_file_type=share->data_file_type) ==
      COMPRESSED_RECORD && testflag & T_UNPACK)
  {
    if (share->options & HA_OPTION_PACK_RECORD)
      sort_info.new_data_file_type = DYNAMIC_RECORD;
    else
      sort_info.new_data_file_type = STATIC_RECORD;
  }

  sort_info.filepos=new_header_length;
  sort_info.dupp=0;
  sort_info.buff=0;
  read_cache.end_of_file=sort_info.filelength=
    my_seek(read_cache.file,0L,MY_SEEK_END,MYF(0));

  if (share->data_file_type == DYNAMIC_RECORD)
    length=max(share->base.min_pack_length+1,share->base.min_block_length);
  else if (share->data_file_type == COMPRESSED_RECORD)
    length=share->base.min_block_length;
  else
    length=share->base.pack_reclength;
  sort_param.max_records=sort_info.max_records=sort_info.filelength/length+1;
  sort_param.key_cmp=sort_key_cmp;
  sort_param.key_write=sort_key_write;
  sort_param.key_read=sort_key_read;
  sort_param.lock_in_memory=lock_memory;
  sort_param.tmpdir=tmpdir;
  del=share->state.del;
  if (sort_info.new_data_file_type != COMPRESSED_RECORD &&
      ! rep_quick)
    share->state.checksum=0;

  rec_per_key_part= info->s->state.rec_per_key_part;
  for (sort_info.key=0 ; sort_info.key < share->base.keys ;
       rec_per_key_part+=sort_info.keyinfo->keysegs, sort_info.key++)
  {
    sort_info.keyinfo=share->keyinfo+sort_info.key;
    if (!(((ulonglong) 1 << sort_info.key) & share->state.key_map))
      continue;

    if ((!(testflag & T_SILENT)))
      printf ("- Fixing index %d\n",sort_info.key+1);
    sort_info.max_pos=sort_info.pos=share->pack.header_length;
    sort_info.keyseg=sort_info.keyinfo->seg;
    sort_info.fix_datafile= (my_bool) (sort_info.key == 0 && ! rep_quick);
    bzero((char*) sort_info.unique,sizeof(sort_info.unique));
    sort_param.key_length=share->rec_reflength;
    for (i=0 ; sort_info.keyseg[i].type != HA_KEYTYPE_END; i++)
    {
      sort_param.key_length+=sort_info.keyseg[i].length;
      if (sort_info.keyseg[i].flag & HA_SPACE_PACK)
	sort_param.key_length+=get_pack_length(sort_info.keyseg[i].length);
      if (sort_info.keyseg[i].flag & (HA_BLOB_PART | HA_VAR_LENGTH))
	sort_param.key_length+=2 + test(sort_info.keyseg[i].length >= 127);
      if (sort_info.keyseg[i].flag & HA_NULL_PART)
	sort_param.key_length++;
    }
    share->state.records=share->state.del=share->state.empty=
      share->state.split=0;

    if (_create_index_by_sort(&sort_param,
			      (my_bool) (!(testflag & T_VERBOSE)),
			      (uint) sort_buffer_length))
      goto err;
		/* Set for next loop */
    sort_param.max_records=sort_info.max_records=share->state.records;

    if (testflag & T_STATISTICS)
      update_key_parts(sort_info.keyinfo, rec_per_key_part, sort_info.unique,
		       (ulonglong) share->state.records);

    if (sort_info.fix_datafile)
    {
      info->dfile=new_file;
      share->state.data_file_length=sort_info.filepos;
      share->state.split=share->state.records; /* Only hole records */
      share->state.version=(ulong) time((time_t*) 0);
      out_flag|=O_NEW_DATA;			/* Data in new file */
      read_cache.end_of_file=sort_info.filepos;
      if (write_data_suffix(info) || end_io_cache(&info->rec_cache))
	goto err;
      share->data_file_type=sort_info.new_data_file_type;
      share->pack.header_length=new_header_length;
    }
    else
      share->state.data_file_length=sort_info.max_pos;

    if (flush_pending_blocks())
      goto err;

    read_cache.file=info->dfile;		/* re-init read cache */
    reinit_io_cache(&read_cache,READ_CACHE,share->pack.header_length,1);
  }

  if (testflag & T_WRITE_LOOP)
  {
    VOID(fputs("          \r",stdout)); VOID(fflush(stdout));
  }

  if (rep_quick && del+sort_info.dupp != share->state.del)
  {
    print_error("Couldn't fix table with quick recovery: Found wrong number of deleted records");
    print_error("Run recovery again without -q");
    got_error=1;
    goto err;
  }

  if (rep_quick != 1)
  {
    my_off_t skr=share->state.data_file_length+
      (share->options & HA_OPTION_COMPRESS_RECORD ?
       MEMMAP_EXTRA_MARGIN : 0);
#ifdef USE_RELOC
    if (share->data_file_type == STATIC_RECORD &&
	skr < share->base.reloc*share->base.min_pack_length)
      skr=share->base.reloc*share->base.min_pack_length;
#endif
    if (skr != sort_info.filelength)
      if (my_chsize(info->dfile,skr,MYF(0)))
	print_warning("Can't change size of datafile,  error: %d",my_errno);
  }
  if (my_chsize(share->kfile,share->state.key_file_length,MYF(0)))
    print_warning("Can't change size of indexfile, error: %d",my_errno);

  if (!(testflag & T_SILENT))
  {
    if (start_records != share->state.records)
      printf("Data records: %s\n", llstr(share->state.records,llbuff));
    if (sort_info.dupp)
      print_warning("%s records have been removed", llstr(sort_info.dupp,llbuff));
  }
  got_error=0;

err:
  if (got_error)
  {
    if (! error_printed)
      print_error("%d when fixing table",my_errno);
    if (new_file >= 0)
    {
      VOID(end_io_cache(&info->rec_cache));
      VOID(my_close(new_file,MYF(0)));
      VOID(my_delete(temp_filename,MYF(MY_WME)));
    }
  }
  my_free((gptr) sort_info.key_block,MYF(MY_ALLOW_ZERO_PTR));
  my_free(sort_info.record,MYF(MY_ALLOW_ZERO_PTR));
  my_free(sort_info.buff,MYF(MY_ALLOW_ZERO_PTR));
  VOID(end_io_cache(&read_cache));
  VOID(end_io_cache(&info->rec_cache));
  info->opt_flag&= ~(READ_CACHE_USED | WRITE_CACHE_USED);
  if (!got_error && testflag & T_UNPACK)
  {
    share->state.header.options[0]&= (uchar) ~HA_OPTION_COMPRESS_RECORD;
    share->pack.header_length=0;
  }
  DBUG_RETURN(got_error);
} /* rep_by_sort */


	/* Read next record and return next key */

static int sort_key_read(void *key)
{
  int error;
  MI_INFO *info;
  DBUG_ENTER("sort_key_read");

  info=sort_info.info;

  if ((error=sort_get_next_record()))
    DBUG_RETURN(error);
  if (info->s->state.records == sort_info.max_records)
  {
    print_error("Found too many records; Can`t continue");
    DBUG_RETURN(1);
  }
  VOID(_mi_make_key(info,sort_info.key,key,sort_info.record,
		    sort_info.filepos));
  DBUG_RETURN(sort_write_record());
} /* sort_key_read */


	/* Read next record from file using parameters in sort_info */
	/* Return -1 if end of file, 0 if ok and > 0 if error */

static int sort_get_next_record(void)
{
  int searching;
  uint found_record,b_type,left_length;
  my_off_t pos;
  byte *to;
  MI_BLOCK_INFO block_info;
  MI_INFO *info;
  MYISAM_SHARE *share;
  char llbuff[22],llbuff2[22];
  DBUG_ENTER("sort_get_next_record");

  info=sort_info.info;
  share=info->s;
  switch (share->data_file_type) {
  case STATIC_RECORD:
    for (;;)
    {
      if (my_b_read(&read_cache,sort_info.record,share->base.pack_reclength))
	DBUG_RETURN(-1);
      sort_info.start_recpos=sort_info.pos;
      if (!sort_info.fix_datafile)
	sort_info.filepos=sort_info.pos;
      sort_info.max_pos=(sort_info.pos+=share->base.pack_reclength);
      share->state.split++;
      if (*sort_info.record)
	DBUG_RETURN(0);
      if (!sort_info.fix_datafile)
      {
	share->state.del++;
	share->state.empty+=share->base.pack_reclength;
      }
    }
  case DYNAMIC_RECORD:
    LINT_INIT(to);
    pos=sort_info.pos;
    searching=(sort_info.fix_datafile && (testflag & T_EXTEND));
    for (;;)
    {
      found_record=block_info.second_read= 0;
      left_length=1;
      if (searching)
	pos=MY_ALIGN(pos,MI_DYN_ALIGN_SIZE);
      do
      {
	if (pos > sort_info.max_pos)
	  sort_info.max_pos=pos;
	if (pos & (MI_DYN_ALIGN_SIZE-1))
	{
	  if (testflag & T_VERBOSE || searching == 0)
	  {
	    print_info("Wrong aligned block at %s",llstr(pos,llbuff));
	    goto try_next;
	  }
	}
	if (found_record && pos == search_after_block)
	  print_info("Block: %s used by record at %s",
		     llstr(search_after_block,llbuff),
		     llstr(sort_info.start_recpos,llbuff2));
	if (_mi_read_cache(&read_cache,(byte*) block_info.header,pos,
			  MI_BLOCK_INFO_HEADER_LENGTH,
			   test(! found_record) | 2))
	{
	  if (found_record)
	  {
	    print_info("Can't read whole record at %s (errno: %d)",
		       llstr(sort_info.start_recpos,llbuff),errno);
	    goto try_next;
	  }
	  DBUG_RETURN(-1);
	}
	if (searching && ! sort_info.fix_datafile)
	{
	  error_printed=1;
	  DBUG_RETURN(1);	/* Something wrong with data */
	}
	if (((b_type=_mi_get_block_info(&block_info,-1,pos)) & BLOCK_ERROR) ||
	    ((b_type & BLOCK_FIRST) &&
	     (block_info.rec_len < (uint) share->base.min_pack_length ||
	      block_info.rec_len > (uint) share->base.max_pack_length)))
	{
	  uint i;
	  if (testflag & T_VERBOSE || searching == 0)
	    print_info("Wrong bytesec: %3d-%3d-%3d at %10s; Skipped",
		       block_info.header[0],block_info.header[1],
		       block_info.header[2],llstr(pos,llbuff));
	  if (found_record)
	    goto try_next;
	  block_info.second_read=0;
	  searching=1;
	  for (i=1 ; i < 11 ; i++) /* Skipp from read string */
	    if (block_info.header[i] >= 1 && block_info.header[i] <= 16)
	      break;
	  pos+=(ulong) i;
	  continue;
	}
	if (b_type & BLOCK_DELETED)
	{
	  bool error=0;
	  if (block_info.block_len+ (uint) (block_info.filepos-pos) <
	      share->base.min_block_length)
	  {
	    if (!searching)
	      print_info("Deleted block with impossible length %u at %s",
			 block_info.block_len,llstr(pos,llbuff));
	    error=1;
	  }
	  else if (block_info.next_filepos != HA_OFFSET_ERROR &&
		   block_info.next_filepos >=
		   info->s->state.data_file_length ||
		   block_info.prev_filepos != HA_OFFSET_ERROR &&
		   block_info.prev_filepos >= info->s->state.data_file_length)
	  {
	    if (!searching)
	      print_info("Delete link points outside datafile at %s",
			 llstr(pos,llbuff));
	    error=1;
	  }
	  if (error)
	  {
	    if (found_record)
	      goto try_next;
	    searching=1;
	    pos++;
	    block_info.second_read=0;
	    continue;
	  }
	}
	else
	{
	  if (block_info.block_len+ (uint) (block_info.filepos-pos) <
	      share->base.min_block_length ||
	      block_info.block_len > (uint) share->base.max_pack_length+
	      MI_SPLIT_LENGTH)
	  {
	    if (!searching)
	      print_info("Found block with impossible length %u at %s; Skipped",
			 block_info.block_len+ (uint) (block_info.filepos-pos),
			 llstr(pos,llbuff));
	    if (found_record)
	      goto try_next;
	    searching=1;
	    pos++;
	    block_info.second_read=0;
	    continue;
	  }
	}
	if (b_type & (BLOCK_DELETED | BLOCK_SYNC_ERROR))
	{
	  if (!sort_info.fix_datafile && (b_type & BLOCK_DELETED))
	  {
	    share->state.empty+=block_info.block_len;
	    share->state.del++;
	    share->state.split++;
	  }
	  if (found_record)
	    goto try_next;
	  if (searching)
	    pos++;
	  else
	    pos=block_info.filepos+block_info.block_len;
	  block_info.second_read=0;
	  continue;
	}

	share->state.split++;
	if (! found_record++)
	{
	  sort_info.find_length=left_length=block_info.rec_len;
	  sort_info.start_recpos=pos;
	  if (!sort_info.fix_datafile)
	    sort_info.filepos=sort_info.start_recpos;
	  if (sort_info.fix_datafile && (testflag & T_EXTEND))
	    sort_info.pos=block_info.filepos+1;
	  else
	    sort_info.pos=block_info.filepos+block_info.block_len;
	  if (share->base.blobs)
	  {
	    if (!(to=mi_fix_rec_buff_for_blob(info,block_info.rec_len)))
	    {
	      print_error("Not enough memory for blob at %s",
			  llstr(sort_info.start_recpos,llbuff));
	      DBUG_RETURN(-1);
	    }
	  }
	  else
	    to= info->rec_buff;
	}
	if (left_length < block_info.data_len || ! block_info.data_len)
	{
	  print_info("Found block with too small length at %s; Skipped",
		     llstr(sort_info.start_recpos,llbuff));
	  goto try_next;
	}
	if (block_info.filepos + block_info.data_len > read_cache.end_of_file)
	{
	  print_info("Found block that points outside data file at %s",
		     llstr(sort_info.start_recpos,llbuff));
	  goto try_next;
	}
	if (_mi_read_cache(&read_cache,to,block_info.filepos,
			   block_info.data_len, test(found_record == 1)))
	{
	  print_info("Read error for block at: %s (error: %d); Skipped",
		     llstr(block_info.filepos,llbuff),my_errno);
	  goto try_next;
	}
	left_length-=block_info.data_len;
	to+=block_info.data_len;
	pos=block_info.next_filepos;
	if (pos == HA_OFFSET_ERROR && left_length)
	{
	  print_info("Wrong block with wrong total length starting at %s",
		     llstr(sort_info.start_recpos,llbuff));
	  goto try_next;
	}
	if (pos + MI_BLOCK_INFO_HEADER_LENGTH > read_cache.end_of_file)
	{
	  print_info("Found link that points at %s (outside data file) at %s",
		     llstr(pos,llbuff2), llstr(sort_info.start_recpos,llbuff));
	  goto try_next;
	}
      } while (left_length);

      if (_mi_rec_unpack(info,sort_info.record,info->rec_buff,
			 sort_info.find_length) != MY_FILE_ERROR)
      {
	if (read_cache.error < 0)
	  DBUG_RETURN(1);
	if ((testflag & (T_EXTEND | T_REP)) || searching)
	{
	  if (info->s->calc_checksum)
	    info->checksum=mi_checksum(info,sort_info.record);
	  if (_mi_rec_check(info, sort_info.record))
	  {
	    print_info("Found wrong packed record at %s",
		       llstr(sort_info.start_recpos,llbuff));
	    goto try_next;
	  }
	}
	DBUG_RETURN(0);
      }
    try_next:
      pos=sort_info.start_recpos+1;
      searching=1;
    }
  case COMPRESSED_RECORD:
    for (searching=0 ;; searching=1, sort_info.pos++)
    {
      if (_mi_read_cache(&read_cache,(byte*) block_info.header,sort_info.pos,
			share->pack.ref_length,1))
	DBUG_RETURN(-1);
      if (searching && ! sort_info.fix_datafile)
      {
	error_printed=1;
	DBUG_RETURN(1);		/* Something wrong with data */
      }
      sort_info.start_recpos=sort_info.pos;
      if (_mi_pack_get_block_info(info,&block_info,-1,sort_info.pos))
	DBUG_RETURN(-1);
      if (!block_info.rec_len &&
	  sort_info.pos + MEMMAP_EXTRA_MARGIN == read_cache.end_of_file)
	DBUG_RETURN(-1);
      if (block_info.rec_len < (uint) share->min_pack_length ||
	  block_info.rec_len > (uint) share->max_pack_length)
      {
	if (! searching)
	  print_info("Found block with wrong recordlength: %d at %s\n",
		     block_info.rec_len, llstr(sort_info.pos,llbuff));
	continue;
      }
      if (_mi_read_cache(&read_cache,(byte*) info->rec_buff,
			block_info.filepos, block_info.rec_len,1))
      {
	if (! searching)
	  print_info("Couldn't read hole record from %s",
		     llstr(sort_info.pos,llbuff));
	continue;
      }
      if (_mi_pack_rec_unpack(info,sort_info.record,info->rec_buff,
			      block_info.rec_len))
      {
	if (! searching)
	  print_info("Found wrong record at %s", llstr(sort_info.pos,llbuff));
	continue;
      }
      if (!sort_info.fix_datafile)
	sort_info.filepos=sort_info.pos;
      sort_info.max_pos=(sort_info.pos=block_info.filepos+
			 block_info.rec_len);
      share->state.split++;
      info->packed_length=block_info.rec_len;
      DBUG_RETURN(0);
    }
  }
  DBUG_RETURN(1);		/* Impossible */
}


	/* Write record to new file */

static int sort_write_record(void)
{
  int flag;
  uint length;
  ulong block_length,reclength;
  byte *from;
  byte block_buff[8];
  MI_INFO *info;
  MYISAM_SHARE *share;
  DBUG_ENTER("sort_write_record");

  info=sort_info.info;
  share=info->s;
  if (sort_info.fix_datafile)
  {
    switch (sort_info.new_data_file_type) {
    case STATIC_RECORD:
      if (my_b_write(&info->rec_cache,sort_info.record,
		     share->base.pack_reclength))
      {
	print_error("%d when writing to datafile",my_errno);
	DBUG_RETURN(1);
      }
      sort_info.filepos+=share->base.pack_reclength;
      info->s->state.checksum+=mi_static_checksum(info, sort_info.record);
      break;
    case DYNAMIC_RECORD:
      if (! info->blobs)
	from=info->rec_buff;
      else
      {
	/* must be sure that local buffer is big enough */
	reclength=info->s->base.pack_reclength+
	  _my_calc_total_blob_length(info,sort_info.record)+
	  ALIGN_SIZE(MI_MAX_DYN_BLOCK_HEADER)+MI_SPLIT_LENGTH+
	  MI_DYN_DELETE_BLOCK_HEADER;
	if (sort_info.buff_length < reclength)
	{
	  if (!(sort_info.buff=my_realloc(sort_info.buff, (uint) reclength,
					  MYF(MY_FREE_ON_ERROR |
					      MY_ALLOW_ZERO_PTR))))
	    DBUG_RETURN(1);
	  sort_info.buff_length=reclength;
	}
	from=sort_info.buff+ALIGN_SIZE(MI_MAX_DYN_BLOCK_HEADER);
      }
      info->checksum=mi_checksum(info,sort_info.record);
      reclength=_mi_rec_pack(info,from,sort_info.record);
      info->s->state.checksum+=info->checksum;
      block_length=reclength+ 3 +test(reclength > 65532L);
      if (block_length < share->base.min_block_length)
	block_length=share->base.min_block_length;
      flag=0;
      info->update|=HA_STATE_WRITE_AT_END;
      block_length=MY_ALIGN(block_length,MI_DYN_ALIGN_SIZE);
      if (_mi_write_part_record(info,0L,block_length,HA_OFFSET_ERROR,
				&from,&reclength,&flag))
      {
	print_error("%d when writing to datafile",my_errno);
	DBUG_RETURN(1);
      }
      sort_info.filepos+=block_length;
      break;
    case COMPRESSED_RECORD:
      reclength=info->packed_length;
      length=save_pack_length(block_buff,reclength);
      if (info->s->base.blobs)
	length+=save_pack_length(block_buff+length,info->blob_length);
      if (my_b_write(&info->rec_cache,block_buff,length) ||
	  my_b_write(&info->rec_cache,(byte*) info->rec_buff,reclength))
      {
	print_error("%d when writing to datafile",my_errno);
	DBUG_RETURN(1);
      }
      sort_info.filepos+=reclength+length;
      break;
    }
  }
  share->state.records++;
  if (testflag & T_WRITE_LOOP && share->state.records % WRITE_COUNT == 0)
  {
    char llbuff[22];
    printf("%s\r", llstr(share->state.records,llbuff)); VOID(fflush(stdout));
  }
  DBUG_RETURN(0);
} /* sort_write_record */


	/* Compare two keys from _create_index_by_sort */

static int sort_key_cmp(const void *a, const void *b)
{
  uint not_used;
  return (_mi_key_cmp(sort_info.keyseg,*((uchar**) a),*((uchar**) b),
		      USE_WHOLE_KEY, SEARCH_SAME,&not_used));
} /* sort_key_cmp */


static int sort_key_write(const void *a)
{
  uint diff_pos;
  char llbuff[22],llbuff2[22];
  int cmp;

  if (sort_info.key_block->inited)
  {
    cmp=_mi_key_cmp(sort_info.keyseg,sort_info.key_block->lastkey,(uchar*) a,
		    USE_WHOLE_KEY,SEARCH_FIND | SEARCH_UPDATE ,&diff_pos);
  }
  else
  {
    cmp= -1;  diff_pos=sort_info.keyinfo->keysegs;
  }
  if ((sort_info.keyinfo->flag & HA_NOSAME) && cmp == 0)
  {
    sort_info.dupp++;
    sort_info.info->lastpos=get_record_for_key(sort_info.info,
					       sort_info.keyinfo,
					       (uchar*) a);
    print_warning("Dupplicate key for record at %10s against record at %10s",
		  llstr(sort_info.info->lastpos,llbuff),
		  llstr(get_record_for_key(sort_info.info,sort_info.keyinfo,
					   sort_info.key_block->lastkey),
			llbuff2));
    if (testflag & T_VERBOSE)
      _mi_print_key(stdout,sort_info.keyseg,(uchar*) a);
    return(sort_delete_record());
  }
  sort_info.unique[diff_pos-1]++;
#ifndef DBUG_OFF
  if (cmp > 0)
  {
    print_error("Fatal intern error: Keys are not in order from sort");
    return(1);
  }
#endif
  return (sort_insert_key(sort_info.key_block,(uchar*) a,HA_OFFSET_ERROR));
} /* sort_key_write */


	/* get pointer to record from a key */

static my_off_t get_record_for_key(MI_INFO *info, MI_KEYDEF *keyinfo,
				   uchar *key)
{
  return _mi_dpos(info,0,key+_mi_keylength(keyinfo,key));
} /* get_record_for_key */


	/* Insert a key in sort-key-blocks */

static int sort_insert_key(register SORT_KEY_BLOCKS *key_block, uchar *key,
			   my_off_t prev_block)
{
  uint a_length,t_length,nod_flag;
  my_off_t filepos;
  uchar *anc_buff,*lastkey;
  MI_KEY_PARAM s_temp;
  MI_INFO *info;
  DBUG_ENTER("sort_insert_key");

  anc_buff=key_block->buff;
  info=sort_info.info;
  lastkey=key_block->lastkey;
  nod_flag= (key_block == sort_info.key_block ? 0 :
	     sort_info.info->s->base.key_reflength);

  if (!key_block->inited)
  {
    key_block->inited=1;
    if (key_block == sort_info.key_block_end)
    {
      print_error("To many key-block-levels; Try increasing sort_key_blocks");
      DBUG_RETURN(1);
    }
    a_length=2+nod_flag;
    key_block->end_pos=anc_buff+2;
    lastkey=0;					/* No previous key in block */
  }
  else
    a_length=mi_getint(anc_buff);

	/* Save pointer to previous block */
  if (nod_flag)
    _mi_kpointer(info,key_block->end_pos,prev_block);

  t_length=(*sort_info.keyinfo->pack_key)(sort_info.keyinfo,nod_flag,
					  (uchar*) 0,lastkey,key,&s_temp);
  (*sort_info.keyinfo->store_key)(sort_info.keyinfo,
				  key_block->end_pos+nod_flag,&s_temp);
  a_length+=t_length;
  mi_putint(anc_buff,a_length,nod_flag);
  key_block->end_pos+=t_length;
  if (a_length <= sort_info.keyinfo->block_length)
  {
    VOID(_mi_move_key(sort_info.keyinfo,key_block->lastkey,key));
    key_block->last_length=a_length-t_length;
    DBUG_RETURN(0);
  }

	/* Fill block with end-zero and write filled block */
  mi_putint(anc_buff,key_block->last_length,nod_flag);
  bzero((byte*) anc_buff+key_block->last_length,
	sort_info.keyinfo->block_length- key_block->last_length);
  if ((filepos=_mi_new(info,sort_info.keyinfo)) == HA_OFFSET_ERROR)
    return 1;
  if (my_pwrite(info->s->kfile,(byte*) anc_buff,
		(uint) sort_info.keyinfo->block_length,filepos,MYF_RW))
    DBUG_RETURN(1);
  DBUG_DUMP("buff",(byte*) anc_buff,mi_getint(anc_buff));

	/* Write separator-key to block in next level */
  if (sort_insert_key(key_block+1,key_block->lastkey,filepos))
    DBUG_RETURN(1);

	/* clear old block and write new key in it */
  key_block->inited=0;
  DBUG_RETURN(sort_insert_key(key_block,key,prev_block));
} /* sort_insert_key */


	/* Delete record when we found a dupplicated key */

static int sort_delete_record(void)
{
  uint i;
  int old_file,error;
  uchar *key;
  MI_INFO *info;
  DBUG_ENTER("sort_delete_record");

  if (opt_rep_quick == 1)
  {
    VOID(fputs("Quick-recover aborted; Run recovery without switch 'q' or with switch -qq\n",stderr));
    error_printed=1;
    DBUG_RETURN(1);
  }
  info=sort_info.info;
  if (info->s->options & HA_OPTION_COMPRESS_RECORD)
  {
    VOID(fputs("Recover aborted; Can't run standard recovery on compressed tables\nwith errors in data-file\nUse switch '--safe-recover' to fix it\n",stderr));
    error_printed=1;
    DBUG_RETURN(1);
  }

  old_file=info->dfile;
  info->dfile=info->rec_cache.file;
  if (sort_info.key)
  {
    key=info->lastkey+info->s->base.max_key_length;
    if ((*info->s->read_rnd)(info,sort_info.record,info->lastpos,0) < 0)
    {
      print_error("Can't read record to be removed");
      info->dfile=old_file;
      DBUG_RETURN(1);
    }

    for (i=0 ; i < sort_info.key ; i++)
    {
      uint key_length=_mi_make_key(info,i,key,sort_info.record,info->lastpos);
      if (_mi_ck_delete(info,i,key,key_length))
      {
	print_error("Can't delete key %d from record to be removed",i+1);
	info->dfile=old_file;
	DBUG_RETURN(1);
      }
    }
    if (info->s->calc_checksum)
      info->s->state.checksum-=(*info->s->calc_checksum)(info,
							 sort_info.record);
  }
  error=flush_io_cache(&info->rec_cache) || (*info->s->delete_record)(info);
  info->dfile=old_file;				/* restore actual value */
  info->s->state.records--;
  DBUG_RETURN(error);
} /* sort_delete_record */


	/* Fix all pending blocks and flush everything to disk */

static int flush_pending_blocks(void)
{
  uint nod_flag,length;
  my_off_t filepos;
  MI_INFO *info;
  SORT_KEY_BLOCKS *key_block;
  DBUG_ENTER("flush_pending_blocks");

  filepos= HA_OFFSET_ERROR;			/* if empty file */
  info=sort_info.info;
  nod_flag=0;
  for (key_block=sort_info.key_block ; key_block->inited ; key_block++)
  {
    key_block->inited=0;
    length=mi_getint(key_block->buff);
    if (nod_flag)
      _mi_kpointer(info,key_block->end_pos,filepos);
    if ((filepos=_mi_new(info,sort_info.keyinfo)) == HA_OFFSET_ERROR)
      DBUG_RETURN(1);
    bzero((byte*) key_block->buff+length,
	  sort_info.keyinfo->block_length-length);
    if (my_pwrite(info->s->kfile,(byte*) key_block->buff,
		 (uint) sort_info.keyinfo->block_length,filepos,MYF_RW))
      DBUG_RETURN(1);
    DBUG_DUMP("buff",(byte*) key_block->buff,length);
    nod_flag=1;
  }
  info->s->state.key_root[sort_info.key]=filepos;	/* Last is root for tree */
  DBUG_RETURN(0);
} /* flush_pending_blocks */


	/* alloc space and pointers for key_blocks */

static SORT_KEY_BLOCKS *alloc_key_blocks(uint blocks, uint buffer_length)
{
  reg1 uint i;
  SORT_KEY_BLOCKS *block;
  DBUG_ENTER("alloc_key_blocks");

  if (!(block=(SORT_KEY_BLOCKS*) my_malloc((sizeof(SORT_KEY_BLOCKS)+
					    buffer_length+IO_SIZE)*blocks,
					   MYF(0))))
  {
    print_error("Not Enough memory for sort-key-blocks");
    return(0);
  }
  for (i=0 ; i < blocks ; i++)
  {
    block[i].inited=0;
    block[i].buff=(uchar*) (block+blocks)+(buffer_length+IO_SIZE)*i;
  }
  DBUG_RETURN(block);
} /* alloc_key_blocks */


	/* print warnings and errors */
	/* VARARGS */

static void print_info(my_string fmt,...)
{
  va_list args;

  va_start(args,fmt);
  VOID(vfprintf(stdout, fmt, args));
  VOID(fputc('\n',stdout));
  va_end(args);
  return;
}

/* VARARGS */

static void print_warning(my_string fmt,...)
{
  va_list args;
  DBUG_ENTER("print_warning");

  fflush(stdout);
  if (!warning_printed && !error_printed)
  {
    if (testflag & T_SILENT)
      fprintf(stderr,"%s: MYISAM file %s\n",my_progname,isam_file_name);
  }
  warning_printed=1;
  va_start(args,fmt);
  fprintf(stderr,"%s: warning: ",my_progname);
  VOID(vfprintf(stderr, fmt, args));
  VOID(fputc('\n',stderr));
  fflush(stderr);
  va_end(args);
  DBUG_VOID_RETURN;
}

/* VARARGS */

void print_error(char *fmt,...)
{
  va_list args;
  DBUG_ENTER("print_error");
  DBUG_PRINT("enter",("format: %s",fmt));

  fflush(stdout);
  if (!warning_printed && !error_printed)
  {
    if (testflag & T_SILENT)
      fprintf(stderr,"%s: ISAM file %s\n",my_progname,isam_file_name);
  }
  error_printed|=1;
  va_start(args,fmt);
  fprintf(stderr,"%s: error: ",my_progname);
  VOID(vfprintf(stderr, fmt, args));
  VOID(fputc('\n',stderr));
  fflush(stderr);
  va_end(args);
  DBUG_VOID_RETURN;
}

	/* Check if file is almost full */

static int test_if_almost_full(MI_INFO *info)
{
  double diff= 0.9;
  if (info->s->options & HA_OPTION_COMPRESS_RECORD)
    return 0;
  return (my_seek(info->s->kfile,0L,MY_SEEK_END,MYF(0)) >
	  (my_off_t) (info->s->base.max_key_file_length*diff) ||
	   my_seek(info->dfile,0L,MY_SEEK_END,MYF(0)) >
	   (my_off_t) (info->s->base.max_data_file_length*diff));
}

	/* Recreate table with bigger more alloced record-data */

static int recreate_database(MI_INFO **org_info, char *filename)
{
  int error;
  MI_INFO info;
  MYISAM_SHARE share;
  MI_KEYDEF *keyinfo,*key,*key_end;
  MI_KEYSEG *keysegs,*keyseg;
  MI_COLUMNDEF *recdef,*rec,*end;
  MI_UNIQUEDEF *uniquedef,*u_ptr,*u_end;
  uint unpack,key_parts;
  ha_rows max_records;
  char name[FN_REFLEN];
  ulonglong file_length,tmp_length;

  error=1;					/* Default error */
  memcpy((byte*) &info,*org_info,sizeof(info));
  memcpy((byte*) &share,(*org_info)->s,sizeof(share));
  unpack= (share.options & HA_OPTION_COMPRESS_RECORD) &&
    (testflag & T_UNPACK);
  if (!(keyinfo=(MI_KEYDEF*) my_alloca(sizeof(MI_KEYDEF)*share.base.keys)))
    return 0;
  memcpy((byte*) keyinfo,(byte*) share.keyinfo,
	 (size_t) (sizeof(MI_KEYDEF)*share.base.keys));

  key_parts= share.base.all_key_parts;
  if (!(keysegs=(MI_KEYSEG*) my_alloca(sizeof(MI_KEYSEG)*
				       (key_parts+share.base.keys))))
  {
    my_afree((gptr) keyinfo);
    return 1;
  }
  if (!(recdef=(MI_COLUMNDEF*)
	my_alloca(sizeof(MI_COLUMNDEF)*(share.base.fields+1))))
  {
    my_afree((gptr) keyinfo);
    my_afree((gptr) keysegs);
    return 1;
  }
  if (!(uniquedef=(MI_UNIQUEDEF*)
	my_alloca(sizeof(MI_UNIQUEDEF)*(share.state.header.uniques+1))))
  {
    my_afree((gptr) recdef);
    my_afree((gptr) keyinfo);
    my_afree((gptr) keysegs);
    return 1;
  }

  /* Copy the column definitions */
  memcpy((byte*) recdef,(byte*) share.rec,
	 (size_t) (sizeof(MI_COLUMNDEF)*(share.base.fields+1)));
  for (rec=recdef,end=recdef+share.base.fields; rec != end ; rec++)
  {
    if (unpack && !(share.options & HA_OPTION_PACK_RECORD) &&
	rec->type != FIELD_BLOB &&
	rec->type != FIELD_VARCHAR &&
	rec->type != FIELD_CHECK)
      rec->type=(int) FIELD_NORMAL;
  }

  /* Change the new key to point at the saved key segments */
  memcpy((byte*) keysegs,(byte*) share.keyparts,
	 (size_t) (sizeof(MI_KEYSEG)*(key_parts+share.base.keys+
				      share.state.header.uniques)));
  keyseg=keysegs;
  for (key=keyinfo,key_end=keyinfo+share.base.keys; key != key_end ; key++)
  {
    key->seg=keyseg;
    keyseg+=key->keysegs+1;
  }
  
  /* Copy the unique definitions and change them to point at the new key
     segments*/
  memcpy((byte*) uniquedef,(byte*) share.uniqueinfo,
	 (size_t) (sizeof(MI_UNIQUEDEF)*(share.state.header.uniques)));
  for (u_ptr=uniquedef,u_end=uniquedef+share.state.header.uniques;
       u_ptr != u_end ; u_ptr++)
  {
    u_ptr->seg=keyseg;
    keyseg+=u_ptr->keysegs+1;
  }
  if (share.options & HA_OPTION_COMPRESS_RECORD)
    share.base.records=max_records=share.state.records;
  else if (share.base.min_pack_length)
    max_records=my_seek(info.dfile,0L,MY_SEEK_END,MYF(0)) /
      (ulong) share.base.min_pack_length;
  else
    max_records=0;
  unpack= (share.options & HA_OPTION_COMPRESS_RECORD) &&
    (testflag & T_UNPACK);
  share.options&= ~HA_OPTION_TEMP_COMPRESS_RECORD;

  file_length=(ulonglong) my_seek(info.dfile,0L,MY_SEEK_END,MYF(0));
  tmp_length= file_length+file_length/10;
  set_if_bigger(file_length,max_data_file_length);
  set_if_bigger(file_length,tmp_length);
  set_if_bigger(file_length,share.base.max_data_file_length);

  VOID(mi_close(*org_info));
  if (mi_create(fn_format(name,filename,"",MI_NAME_IEXT,
			  4+ (opt_follow_links ? 16 : 0)),
		share.base.keys - share.state.header.uniques,
		keyinfo, share.base.fields, recdef,
		share.state.header.uniques, uniquedef,
		max(max_records,share.base.records),share.base.reloc,
		HA_DONT_TOUCH_DATA,
		share.options | (unpack ? HA_OPTION_TEMP_COMPRESS_RECORD
				     : 0),
		file_length,
		share.state.auto_increment))
  {
    print_error("Got error %d when trying to recreate indexfile",my_errno);
    goto end;
  }
  *org_info=mi_open(name,O_RDWR,
		    (testflag & T_WAIT_FOREVER) ? HA_OPEN_WAIT_IF_LOCKED :
		    (testflag & T_DESCRIPT) ? HA_OPEN_IGNORE_IF_LOCKED :
		    HA_OPEN_ABORT_IF_LOCKED);
  if (!*org_info)
  {
    print_error("Got error %d when trying to open re-created indexfile",
		my_errno);
    goto end;
  }
  /* We are modifing */
  (*org_info)->s->options&= ~HA_OPTION_READ_ONLY_DATA;
  VOID(_mi_readinfo(*org_info,F_WRLCK,0));
  (*org_info)->s->state.records=share.state.records;
  if (share.state.create_time)
    (*org_info)->s->state.create_time=share.state.create_time;
  (*org_info)->s->state.unique=(*org_info)->this_unique=
    share.state.unique;
  (*org_info)->s->state.checksum=share.state.checksum;
  (*org_info)->s->state.del=share.state.del;
  (*org_info)->s->state.dellink=share.state.dellink;
  (*org_info)->s->state.empty=share.state.empty;
  (*org_info)->s->state.data_file_length=share.state.data_file_length;
  if (update_state_info(*org_info,UPDATE_TIME | UPDATE_STAT))
    goto end;
  error=0;
end:
  my_afree((gptr) uniquedef);
  my_afree((gptr) keyinfo);
  my_afree((gptr) recdef);
  my_afree((gptr) keysegs);
  return error;
}


	/* write suffix to data file if neaded */

static int write_data_suffix(MI_INFO *info)
{
  if (info->s->options & HA_OPTION_COMPRESS_RECORD &&
      sort_info.fix_datafile)
  {
    char buff[MEMMAP_EXTRA_MARGIN];
    bzero(buff,sizeof(buff));
    if (my_b_write(&info->rec_cache,buff,sizeof(buff)))
    {
      print_error("%d when writing to datafile",my_errno);
      return 1;
    }
    read_cache.end_of_file+=sizeof(buff);
  }
  return 0;
}


	/* Update state and myisamchk_time of indexfile */

static int update_state_info(MI_INFO *info,uint update)
{
  MYISAM_SHARE *share=info->s;

  if (update != UPDATE_AUTO_INC)
    share->state.open_count=0;
  if (update & UPDATE_STAT)
    share->state.rec_per_key_rows=share->state.records;
  if (update & (UPDATE_STAT | UPDATE_SORT | UPDATE_TIME | UPDATE_AUTO_INC))
  {
    if (update & UPDATE_TIME)
    {
      share->state.check_time= (long) time((time_t*) 0);
      if (!share->state.create_time)
	share->state.create_time=share->state.check_time;
    }
    if (mi_state_info_write(share->kfile,&share->state,1+2))
      goto err;
  }
  {						/* Force update of status */
    int error;
    uint r_locks=share->r_locks,w_locks=share->w_locks;
    share->r_locks=share->w_locks=0;
    error=_mi_writeinfo(info,WRITEINFO_NO_UNLOCK);
    share->r_locks=r_locks; share->w_locks=w_locks;
    if (!error)
      return 0;
  }
err:
  print_error("%d when updateing keyfile",my_errno);
  return 1;
}


char *llstr(longlong value,char *buff)
{
  longlong2str(value,buff,10);
  return buff;
}

	/* Update auto increment value for a table */

static void update_auto_increment_key(MI_INFO *info)
{
  if (!info->s->base.auto_key ||
      !(((ulonglong) 1 << (info->s->base.auto_key-1)
	 & info->s->state.key_map)))
  {
    printf("Table: %s doesn't have an auto increment key\n",
	   isam_file_name);
    return;
  }
  if (!(testflag & T_SILENT))
    printf("Updating MYISAM file: %s\n", isam_file_name);
  if (mi_rlast(info,info->rec_buff, info->s->base.auto_key-1))
  {
    if (errno != HA_ERR_END_OF_FILE)
    {
      print_error("%d when reading last record",errno);
      return;
    }
    info->s->state.auto_increment=auto_increment_value;
  }
  else
  {
    info->s->state.auto_increment=0;
    update_auto_increment(info,info->rec_buff);
    set_if_bigger(info->s->state.auto_increment,auto_increment_value)
  }
  update_state_info(info, UPDATE_AUTO_INC);
  return;
}

    /* calculate unique keys for each part key */

static void update_key_parts(MI_KEYDEF *keyinfo,
			     ulong *rec_per_key_part,
			     ulonglong *unique,
			     ulonglong records)
{
  ulonglong count=0,tmp;
  uint parts;
  for (parts=0 ; parts < keyinfo->keysegs  ; parts++)
  {
    count+=unique[parts];
    if (count == 0)
      tmp=records;
    else
      tmp= (records+count/2) / count;
    if (tmp >= (ulonglong) ~(ulong) 0)
      tmp=(ulonglong) ~(ulong) 0;
    *rec_per_key_part=(ulong) tmp;
    rec_per_key_part++;
  }
}


ha_checksum mi_byte_checksum(const byte *buf, uint length)
{
  ha_checksum crc;
  const byte *end=buf+length;
  for (crc=0; buf != end; buf++)
    crc=((crc << 1) + *((uchar*) buf)) +
      test(crc & (((ha_checksum) 1) << (8*sizeof(ha_checksum)-1)));
  return crc;
}
