/* Copyright (C) 1979-1996 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.

   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* When using sql procedures */

#ifdef __GNUC__
#pragma interface				/* gcc class implementation */
#endif

#define PROC_NO_SORT 1				/* Bits in flags */
#define PROC_GROUP   2				/* proc must have group */

/* Procedure items used by procedures to store values for send_fields */

class Item_proc :public Item
{
public:
  Item_proc(char *name_par): Item()
  {
     this->name=name_par;
  }
  enum Type type() const { return Item::PROC_ITEM; }
  virtual void set(double nr)=0;
  virtual void set(const char *str,uint length)=0;
  virtual void set(longlong nr)=0;
  virtual enum_field_types field_type() const=0;
  void set(const char *str) { set(str,strlen(str)); }
  void make_field(Send_field *tmp_field)
  {
    init_make_field(tmp_field,field_type());
  }
};

class Item_proc_real :public Item_proc
{
  double value;
public:
  Item_proc_real(char *name_par,uint dec) : Item_proc(name_par)
  {
     decimals=dec; max_length=float_length(dec); 
  }
  enum Item_result result_type () const { return REAL_RESULT; }
  enum_field_types field_type() const { return FIELD_TYPE_DOUBLE; }
  void set(double nr) { value=nr; }
  void set(longlong nr) { value=(double) nr; }
  void set(const char *str,uint length) { value=atof(str); }
  double val() { return value; }
  longlong val_int() { return (longlong) value; }
  String *str(String *s) { s->set(value,decimals); return s; }
};

class Item_proc_int :public Item_proc
{
  longlong value;
public:
  Item_proc_int(char *name_par) :Item_proc(name_par)
  { max_length=11; }
  enum Item_result result_type () const { return INT_RESULT; }
  enum_field_types field_type() const { return FIELD_TYPE_LONG; }
  void set(double nr) { value=(longlong) nr; }
  void set(longlong nr) { value=nr; }
  void set(const char *str,uint length) { value=strtoll(str,NULL,10); }
  double val() { return (double) value; }
  longlong val_int() { return value; }
  String *str(String *s) { s->set(value); return s; }
};


class Item_proc_string :public Item_proc
{
public:
  Item_proc_string(char *name_par,uint length) :Item_proc(name_par)
    { this->max_length=length; }
  enum Item_result result_type () const { return STRING_RESULT; }
  enum_field_types field_type() const { return FIELD_TYPE_STRING; }
  void set(double nr) { str_value.set(nr); }
  void set(longlong nr) { str_value.set(nr); }
  void set(const char *str, uint length) { str_value.copy(str,length); }
  double val() { return atof(str_value.ptr()); }
  longlong val_int() { return strtoll(str_value.ptr(),NULL,10); }
  String *str(String*) 
  { 
    return null_value ? (String*) 0 : (String*) &str_value; 
  }
};

/* The procedure class definitions */

class Procedure {
protected:
  List<Item> *fields;
  select_result *result;
public:
  const uint flags;
  ORDER *group,*param_fields;
  Procedure(select_result *res,uint flags_par) :result(res),flags(flags_par),
    group(0),param_fields(0) {}
  virtual ~Procedure() {group=param_fields=0; fields=0; }
  virtual void add(void)=0;
  virtual void end_group(void)=0;
  virtual int send_row(List<Item> &fields)=0;
  virtual bool change_columns(List<Item> &fields)=0;
  virtual bool end_of_records() { return 0; }
};

Procedure *setup_procedure(THD *thd,ORDER *proc_param,select_result *result,
			   List<Item> &field_list,int *error);
