/* Copyright (C) 1998 TcX AB & Monty Program KB & Detron HB &
   Alexis Mikhailov <root@medinf.chuvashia.su>

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.

   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* This file defines structures needed by udf functions */

#ifdef __GNUC__
#pragma interface
#endif

enum Item_udftype {UDFTYPE_FUNCTION=1,UDFTYPE_AGGREGATE};

typedef struct st_udf_func
{
  char *name;
  int name_length;
  Item_result returns;
  Item_udftype type;
  char *dl;
  void *dlhandle;
  void *func;
  void *func_init;
  void *func_deinit;
  void *func_reset;
  void *func_add;
  ulong usage_count;
} udf_func;

class Item_result_field;
struct st_table_list;

class udf_handler :public Sql_alloc
{
 protected:
  udf_func *u_d;
  String *buffers;
  UDF_ARGS f_args;
  UDF_INIT initid;
  char *num_buffer;
  uchar error;
  bool initialized;
  Item **args;

 public:
  table_map used_tables_cache;
  udf_handler(udf_func *udf_arg) :u_d(udf_arg), buffers(0), error(0),
    initialized(0)
  {}
  ~udf_handler();
  const char *name() const { return u_d ? u_d->name : "?"; }
  Item_result result_type () const
  { return u_d  ? u_d->returns : STRING_RESULT;}
  bool get_arguments();
  bool fix_fields(THD *thd,struct st_table_list *tlist,Item_result_field *item,
		  uint arg_count,Item **args);
  double val(my_bool *null_value)
  {
    if (get_arguments())
    {
      *null_value=1;
      return 0.0;
    }
    uchar is_null=0;
    double (*func)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)=
      (double (*)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)) u_d->func;
    double tmp=func(&initid, &f_args, &is_null, &error);
    if (is_null || error)
    {
      *null_value=1;
      return 0.0;
    }
    *null_value=0;
    return tmp;
  }
  longlong val_int(my_bool *null_value)
  {
    if (get_arguments())
    {
      *null_value=1;
      return 0LL;
    }
    uchar is_null=0;
    longlong (*func)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)=
      (longlong (*)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)) u_d->func;
    longlong tmp=func(&initid, &f_args, &is_null, &error);
    if (is_null || error)
    {
      *null_value=1;
      return 0LL;
    }
    *null_value=0;
    return tmp;
  }
  void reset(my_bool *null_value)
  {
    uchar is_null=0;
    if (get_arguments())
    {
      *null_value=1;
      return;
    }
    void (*func)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)=
    (void (*)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)) u_d->func_reset;
    func(&initid, &f_args, &is_null, &error);
    *null_value= (my_bool) (is_null || error);
  }
  void add(my_bool *null_value)
  {
    uchar is_null=0;
    if (get_arguments())
    {
      *null_value=1;
      return;
    }
    void (*func)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)=
    (void (*)(UDF_INIT *, UDF_ARGS *, uchar *, uchar *)) u_d->func_add;
    func(&initid, &f_args, &is_null, &error);
    *null_value= (my_bool) (is_null || error);
  }
  String *str(String *str,String *save_str);
};


#ifdef HAVE_DLOPEN
void udf_init(void),udf_free(void);
udf_func *find_udf(const char *name, uint len=0,bool mark_used=0);
void free_udf(udf_func *udf);
int mysql_create_function(THD *thd,udf_func *udf);
int mysql_drop_function(THD *thd,const char *name);
#endif
